/*
 * Copyright 2007-8 Advanced Micro Devices, Inc.
 * Copyright 2008 Red Hat Inc.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE COPYRIGHT HOLDER(S) OR AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * Authors: Dave Airlie
 *          Alex Deucher
 */
#include "drmP.h"
#include "radeon_drm.h"
#include "radeon.h"

#include "atom.h"
#include "atom-bits.h"

/* from radeon_encoder.c */
extern uint32_t
radeon_get_encoder_id(struct drm_device *dev, uint32_t supported_device,
		      uint8_t dac);
extern void radeon_link_encoder_connector(struct drm_device *dev);
extern void
radeon_add_atom_encoder(struct drm_device *dev, uint32_t encoder_id,
			uint32_t supported_device);

/* from radeon_connector.c */
extern void
radeon_add_atom_connector(struct drm_device *dev,
			  uint32_t connector_id,
			  uint32_t supported_device,
			  int connector_type,
			  struct radeon_i2c_bus_rec *i2c_bus,
			  bool linkb, uint32_t igp_lane_info,
			  uint16_t connector_object_id,
			  struct radeon_hpd *hpd);

/* from radeon_legacy_encoder.c */
extern void
radeon_add_legacy_encoder(struct drm_device *dev, uint32_t encoder_id,
			  uint32_t supported_device);

union atom_supported_devices {
	struct _ATOM_SUPPORTED_DEVICES_INFO info;
	struct _ATOM_SUPPORTED_DEVICES_INFO_2 info_2;
	struct _ATOM_SUPPORTED_DEVICES_INFO_2d1 info_2d1;
};

static inline struct radeon_i2c_bus_rec radeon_lookup_i2c_gpio(struct radeon_device *rdev,
							       uint8_t id)
{
	struct atom_context *ctx = rdev-&gt;mode_info.atom_context;
	ATOM_GPIO_I2C_ASSIGMENT *gpio;
	struct radeon_i2c_bus_rec i2c;
	int index = GetIndexIntoMasterTable(DATA, GPIO_I2C_Info);
	struct _ATOM_GPIO_I2C_INFO *i2c_info;
	uint16_t data_offset, size;
	int i, num_indices;

	memset(&amp;i2c, 0, sizeof(struct radeon_i2c_bus_rec));
	i2c.valid = false;

	if (atom_parse_data_header(ctx, index, &amp;size, NULL, NULL, &amp;data_offset)) {
		i2c_info = (struct _ATOM_GPIO_I2C_INFO *)(ctx-&gt;bios + data_offset);

		num_indices = (size - sizeof(ATOM_COMMON_TABLE_HEADER)) /
			sizeof(ATOM_GPIO_I2C_ASSIGMENT);

		for (i = 0; i &lt; num_indices; i++) {
			gpio = &amp;i2c_info-&gt;asGPIO_Info[i];

			if (gpio-&gt;sucI2cId.ucAccess == id) {
				i2c.mask_clk_reg = le16_to_cpu(gpio-&gt;usClkMaskRegisterIndex) * 4;
				i2c.mask_data_reg = le16_to_cpu(gpio-&gt;usDataMaskRegisterIndex) * 4;
				i2c.en_clk_reg = le16_to_cpu(gpio-&gt;usClkEnRegisterIndex) * 4;
				i2c.en_data_reg = le16_to_cpu(gpio-&gt;usDataEnRegisterIndex) * 4;
				i2c.y_clk_reg = le16_to_cpu(gpio-&gt;usClkY_RegisterIndex) * 4;
				i2c.y_data_reg = le16_to_cpu(gpio-&gt;usDataY_RegisterIndex) * 4;
				i2c.a_clk_reg = le16_to_cpu(gpio-&gt;usClkA_RegisterIndex) * 4;
				i2c.a_data_reg = le16_to_cpu(gpio-&gt;usDataA_RegisterIndex) * 4;
				i2c.mask_clk_mask = (1 &lt;&lt; gpio-&gt;ucClkMaskShift);
				i2c.mask_data_mask = (1 &lt;&lt; gpio-&gt;ucDataMaskShift);
				i2c.en_clk_mask = (1 &lt;&lt; gpio-&gt;ucClkEnShift);
				i2c.en_data_mask = (1 &lt;&lt; gpio-&gt;ucDataEnShift);
				i2c.y_clk_mask = (1 &lt;&lt; gpio-&gt;ucClkY_Shift);
				i2c.y_data_mask = (1 &lt;&lt; gpio-&gt;ucDataY_Shift);
				i2c.a_clk_mask = (1 &lt;&lt; gpio-&gt;ucClkA_Shift);
				i2c.a_data_mask = (1 &lt;&lt; gpio-&gt;ucDataA_Shift);

				if (gpio-&gt;sucI2cId.sbfAccess.bfHW_Capable)
					i2c.hw_capable = true;
				else
					i2c.hw_capable = false;

				if (gpio-&gt;sucI2cId.ucAccess == 0xa0)
					i2c.mm_i2c = true;
				else
					i2c.mm_i2c = false;

				i2c.i2c_id = gpio-&gt;sucI2cId.ucAccess;

				i2c.valid = true;
				break;
			}
		}
	}

	return i2c;
}

static inline struct radeon_gpio_rec radeon_lookup_gpio(struct radeon_device *rdev,
							u8 id)
{
	struct atom_context *ctx = rdev-&gt;mode_info.atom_context;
	struct radeon_gpio_rec gpio;
	int index = GetIndexIntoMasterTable(DATA, GPIO_Pin_LUT);
	struct _ATOM_GPIO_PIN_LUT *gpio_info;
	ATOM_GPIO_PIN_ASSIGNMENT *pin;
	u16 data_offset, size;
	int i, num_indices;

	memset(&amp;gpio, 0, sizeof(struct radeon_gpio_rec));
	gpio.valid = false;

	if (atom_parse_data_header(ctx, index, &amp;size, NULL, NULL, &amp;data_offset)) {
		gpio_info = (struct _ATOM_GPIO_PIN_LUT *)(ctx-&gt;bios + data_offset);

		num_indices = (size - sizeof(ATOM_COMMON_TABLE_HEADER)) /
			sizeof(ATOM_GPIO_PIN_ASSIGNMENT);

		for (i = 0; i &lt; num_indices; i++) {
			pin = &amp;gpio_info-&gt;asGPIO_Pin[i];
			if (id == pin-&gt;ucGPIO_ID) {
				gpio.id = pin-&gt;ucGPIO_ID;
				gpio.reg = pin-&gt;usGpioPin_AIndex * 4;
				gpio.mask = (1 &lt;&lt; pin-&gt;ucGpioPinBitShift);
				gpio.valid = true;
				break;
			}
		}
	}

	return gpio;
}

static struct radeon_hpd radeon_atom_get_hpd_info_from_gpio(struct radeon_device *rdev,
							    struct radeon_gpio_rec *gpio)
{
	struct radeon_hpd hpd;
	u32 reg;

	if (ASIC_IS_DCE4(rdev))
		reg = EVERGREEN_DC_GPIO_HPD_A;
	else
		reg = AVIVO_DC_GPIO_HPD_A;

	hpd.gpio = *gpio;
	if (gpio-&gt;reg == reg) {
		switch(gpio-&gt;mask) {
		case (1 &lt;&lt; 0):
			hpd.hpd = RADEON_HPD_1;
			break;
		case (1 &lt;&lt; 8):
			hpd.hpd = RADEON_HPD_2;
			break;
		case (1 &lt;&lt; 16):
			hpd.hpd = RADEON_HPD_3;
			break;
		case (1 &lt;&lt; 24):
			hpd.hpd = RADEON_HPD_4;
			break;
		case (1 &lt;&lt; 26):
			hpd.hpd = RADEON_HPD_5;
			break;
		case (1 &lt;&lt; 28):
			hpd.hpd = RADEON_HPD_6;
			break;
		default:
			hpd.hpd = RADEON_HPD_NONE;
			break;
		}
	} else
		hpd.hpd = RADEON_HPD_NONE;
	return hpd;
}

static bool radeon_atom_apply_quirks(struct drm_device *dev,
				     uint32_t supported_device,
				     int *connector_type,
				     struct radeon_i2c_bus_rec *i2c_bus,
				     uint16_t *line_mux,
				     struct radeon_hpd *hpd)
{

	/* Asus M2A-VM HDMI board lists the DVI port as HDMI */
	if ((dev-&gt;pdev-&gt;device == 0x791e) &amp;&amp;
	    (dev-&gt;pdev-&gt;subsystem_vendor == 0x1043) &amp;&amp;
	    (dev-&gt;pdev-&gt;subsystem_device == 0x826d)) {
		if ((*connector_type == DRM_MODE_CONNECTOR_HDMIA) &amp;&amp;
		    (supported_device == ATOM_DEVICE_DFP3_SUPPORT))
			*connector_type = DRM_MODE_CONNECTOR_DVID;
	}

	/* Asrock RS600 board lists the DVI port as HDMI */
	if ((dev-&gt;pdev-&gt;device == 0x7941) &amp;&amp;
	    (dev-&gt;pdev-&gt;subsystem_vendor == 0x1849) &amp;&amp;
	    (dev-&gt;pdev-&gt;subsystem_device == 0x7941)) {
		if ((*connector_type == DRM_MODE_CONNECTOR_HDMIA) &amp;&amp;
		    (supported_device == ATOM_DEVICE_DFP3_SUPPORT))
			*connector_type = DRM_MODE_CONNECTOR_DVID;
	}

	/* a-bit f-i90hd - ciaranm on #radeonhd - this board has no DVI */
	if ((dev-&gt;pdev-&gt;device == 0x7941) &amp;&amp;
	    (dev-&gt;pdev-&gt;subsystem_vendor == 0x147b) &amp;&amp;
	    (dev-&gt;pdev-&gt;subsystem_device == 0x2412)) {
		if (*connector_type == DRM_MODE_CONNECTOR_DVII)
			return false;
	}

	/* Falcon NW laptop lists vga ddc line for LVDS */
	if ((dev-&gt;pdev-&gt;device == 0x5653) &amp;&amp;
	    (dev-&gt;pdev-&gt;subsystem_vendor == 0x1462) &amp;&amp;
	    (dev-&gt;pdev-&gt;subsystem_device == 0x0291)) {
		if (*connector_type == DRM_MODE_CONNECTOR_LVDS) {
			i2c_bus-&gt;valid = false;
			*line_mux = 53;
		}
	}

	/* HIS X1300 is DVI+VGA, not DVI+DVI */
	if ((dev-&gt;pdev-&gt;device == 0x7146) &amp;&amp;
	    (dev-&gt;pdev-&gt;subsystem_vendor == 0x17af) &amp;&amp;
	    (dev-&gt;pdev-&gt;subsystem_device == 0x2058)) {
		if (supported_device == ATOM_DEVICE_DFP1_SUPPORT)
			return false;
	}

	/* Gigabyte X1300 is DVI+VGA, not DVI+DVI */
	if ((dev-&gt;pdev-&gt;device == 0x7142) &amp;&amp;
	    (dev-&gt;pdev-&gt;subsystem_vendor == 0x1458) &amp;&amp;
	    (dev-&gt;pdev-&gt;subsystem_device == 0x2134)) {
		if (supported_device == ATOM_DEVICE_DFP1_SUPPORT)
			return false;
	}


	/* Funky macbooks */
	if ((dev-&gt;pdev-&gt;device == 0x71C5) &amp;&amp;
	    (dev-&gt;pdev-&gt;subsystem_vendor == 0x106b) &amp;&amp;
	    (dev-&gt;pdev-&gt;subsystem_device == 0x0080)) {
		if ((supported_device == ATOM_DEVICE_CRT1_SUPPORT) ||
		    (supported_device == ATOM_DEVICE_DFP2_SUPPORT))
			return false;
		if (supported_device == ATOM_DEVICE_CRT2_SUPPORT)
			*line_mux = 0x90;
	}

	/* ASUS HD 3600 XT board lists the DVI port as HDMI */
	if ((dev-&gt;pdev-&gt;device == 0x9598) &amp;&amp;
	    (dev-&gt;pdev-&gt;subsystem_vendor == 0x1043) &amp;&amp;
	    (dev-&gt;pdev-&gt;subsystem_device == 0x01da)) {
		if (*connector_type == DRM_MODE_CONNECTOR_HDMIA) {
			*connector_type = DRM_MODE_CONNECTOR_DVII;
		}
	}

	/* ASUS HD 3450 board lists the DVI port as HDMI */
	if ((dev-&gt;pdev-&gt;device == 0x95C5) &amp;&amp;
	    (dev-&gt;pdev-&gt;subsystem_vendor == 0x1043) &amp;&amp;
	    (dev-&gt;pdev-&gt;subsystem_device == 0x01e2)) {
		if (*connector_type == DRM_MODE_CONNECTOR_HDMIA) {
			*connector_type = DRM_MODE_CONNECTOR_DVII;
		}
	}

	/* some BIOSes seem to report DAC on HDMI - usually this is a board with
	 * HDMI + VGA reporting as HDMI
	 */
	if (*connector_type == DRM_MODE_CONNECTOR_HDMIA) {
		if (supported_device &amp; (ATOM_DEVICE_CRT_SUPPORT)) {
			*connector_type = DRM_MODE_CONNECTOR_VGA;
			*line_mux = 0;
		}
	}

	/* Acer laptop reports DVI-D as DVI-I */
	if ((dev-&gt;pdev-&gt;device == 0x95c4) &amp;&amp;
	    (dev-&gt;pdev-&gt;subsystem_vendor == 0x1025) &amp;&amp;
	    (dev-&gt;pdev-&gt;subsystem_device == 0x013c)) {
		if ((*connector_type == DRM_MODE_CONNECTOR_DVII) &amp;&amp;
		    (supported_device == ATOM_DEVICE_DFP1_SUPPORT))
			*connector_type = DRM_MODE_CONNECTOR_DVID;
	}

	/* XFX Pine Group device rv730 reports no VGA DDC lines
	 * even though they are wired up to record 0x93
	 */
	if ((dev-&gt;pdev-&gt;device == 0x9498) &amp;&amp;
	    (dev-&gt;pdev-&gt;subsystem_vendor == 0x1682) &amp;&amp;
	    (dev-&gt;pdev-&gt;subsystem_device == 0x2452)) {
		struct radeon_device *rdev = dev-&gt;dev_private;
		*i2c_bus = radeon_lookup_i2c_gpio(rdev, 0x93);
	}
	return true;
}

const int supported_devices_connector_convert[] = {
	DRM_MODE_CONNECTOR_Unknown,
	DRM_MODE_CONNECTOR_VGA,
	DRM_MODE_CONNECTOR_DVII,
	DRM_MODE_CONNECTOR_DVID,
	DRM_MODE_CONNECTOR_DVIA,
	DRM_MODE_CONNECTOR_SVIDEO,
	DRM_MODE_CONNECTOR_Composite,
	DRM_MODE_CONNECTOR_LVDS,
	DRM_MODE_CONNECTOR_Unknown,
	DRM_MODE_CONNECTOR_Unknown,
	DRM_MODE_CONNECTOR_HDMIA,
	DRM_MODE_CONNECTOR_HDMIB,
	DRM_MODE_CONNECTOR_Unknown,
	DRM_MODE_CONNECTOR_Unknown,
	DRM_MODE_CONNECTOR_9PinDIN,
	DRM_MODE_CONNECTOR_DisplayPort
};

const uint16_t supported_devices_connector_object_id_convert[] = {
	CONNECTOR_OBJECT_ID_NONE,
	CONNECTOR_OBJECT_ID_VGA,
	CONNECTOR_OBJECT_ID_DUAL_LINK_DVI_I, /* not all boards support DL */
	CONNECTOR_OBJECT_ID_DUAL_LINK_DVI_D, /* not all boards support DL */
	CONNECTOR_OBJECT_ID_VGA, /* technically DVI-A */
	CONNECTOR_OBJECT_ID_COMPOSITE,
	CONNECTOR_OBJECT_ID_SVIDEO,
	CONNECTOR_OBJECT_ID_LVDS,
	CONNECTOR_OBJECT_ID_9PIN_DIN,
	CONNECTOR_OBJECT_ID_9PIN_DIN,
	CONNECTOR_OBJECT_ID_DISPLAYPORT,
	CONNECTOR_OBJECT_ID_HDMI_TYPE_A,
	CONNECTOR_OBJECT_ID_HDMI_TYPE_B,
	CONNECTOR_OBJECT_ID_SVIDEO
};

const int object_connector_convert[] = {
	DRM_MODE_CONNECTOR_Unknown,
	DRM_MODE_CONNECTOR_DVII,
	DRM_MODE_CONNECTOR_DVII,
	DRM_MODE_CONNECTOR_DVID,
	DRM_MODE_CONNECTOR_DVID,
	DRM_MODE_CONNECTOR_VGA,
	DRM_MODE_CONNECTOR_Composite,
	DRM_MODE_CONNECTOR_SVIDEO,
	DRM_MODE_CONNECTOR_Unknown,
	DRM_MODE_CONNECTOR_Unknown,
	DRM_MODE_CONNECTOR_9PinDIN,
	DRM_MODE_CONNECTOR_Unknown,
	DRM_MODE_CONNECTOR_HDMIA,
	DRM_MODE_CONNECTOR_HDMIB,
	DRM_MODE_CONNECTOR_LVDS,
	DRM_MODE_CONNECTOR_9PinDIN,
	DRM_MODE_CONNECTOR_Unknown,
	DRM_MODE_CONNECTOR_Unknown,
	DRM_MODE_CONNECTOR_Unknown,
	DRM_MODE_CONNECTOR_DisplayPort,
	DRM_MODE_CONNECTOR_eDP,
	DRM_MODE_CONNECTOR_Unknown
};

bool radeon_get_atom_connector_info_from_object_table(struct drm_device *dev)
{
	struct radeon_device *rdev = dev-&gt;dev_private;
	struct radeon_mode_info *mode_info = &amp;rdev-&gt;mode_info;
	struct atom_context *ctx = mode_info-&gt;atom_context;
	int index = GetIndexIntoMasterTable(DATA, Object_Header);
	u16 size, data_offset;
	u8 frev, crev;
	ATOM_CONNECTOR_OBJECT_TABLE *con_obj;
	ATOM_DISPLAY_OBJECT_PATH_TABLE *path_obj;
	ATOM_OBJECT_HEADER *obj_header;
	int i, j, path_size, device_support;
	int connector_type;
	u16 igp_lane_info, conn_id, connector_object_id;
	bool linkb;
	struct radeon_i2c_bus_rec ddc_bus;
	struct radeon_gpio_rec gpio;
	struct radeon_hpd hpd;

	if (!atom_parse_data_header(ctx, index, &amp;size, &amp;frev, &amp;crev, &amp;data_offset))
		return false;

	if (crev &lt; 2)
		return false;

	obj_header = (ATOM_OBJECT_HEADER *) (ctx-&gt;bios + data_offset);
	path_obj = (ATOM_DISPLAY_OBJECT_PATH_TABLE *)
	    (ctx-&gt;bios + data_offset +
	     le16_to_cpu(obj_header-&gt;usDisplayPathTableOffset));
	con_obj = (ATOM_CONNECTOR_OBJECT_TABLE *)
	    (ctx-&gt;bios + data_offset +
	     le16_to_cpu(obj_header-&gt;usConnectorObjectTableOffset));
	device_support = le16_to_cpu(obj_header-&gt;usDeviceSupport);

	path_size = 0;
	for (i = 0; i &lt; path_obj-&gt;ucNumOfDispPath; i++) {
		uint8_t *addr = (uint8_t *) path_obj-&gt;asDispPath;
		ATOM_DISPLAY_OBJECT_PATH *path;
		addr += path_size;
		path = (ATOM_DISPLAY_OBJECT_PATH *) addr;
		path_size += le16_to_cpu(path-&gt;usSize);
		linkb = false;
		if (device_support &amp; le16_to_cpu(path-&gt;usDeviceTag)) {
			uint8_t con_obj_id, con_obj_num, con_obj_type;

			con_obj_id =
			    (le16_to_cpu(path-&gt;usConnObjectId) &amp; OBJECT_ID_MASK)
			    &gt;&gt; OBJECT_ID_SHIFT;
			con_obj_num =
			    (le16_to_cpu(path-&gt;usConnObjectId) &amp; ENUM_ID_MASK)
			    &gt;&gt; ENUM_ID_SHIFT;
			con_obj_type =
			    (le16_to_cpu(path-&gt;usConnObjectId) &amp;
			     OBJECT_TYPE_MASK) &gt;&gt; OBJECT_TYPE_SHIFT;

			/* TODO CV support */
			if (le16_to_cpu(path-&gt;usDeviceTag) ==
				ATOM_DEVICE_CV_SUPPORT)
				continue;

			/* IGP chips */
			if ((rdev-&gt;flags &amp; RADEON_IS_IGP) &amp;&amp;
			    (con_obj_id ==
			     CONNECTOR_OBJECT_ID_PCIE_CONNECTOR)) {
				uint16_t igp_offset = 0;
				ATOM_INTEGRATED_SYSTEM_INFO_V2 *igp_obj;

				index =
				    GetIndexIntoMasterTable(DATA,
							    IntegratedSystemInfo);

				if (atom_parse_data_header(ctx, index, &amp;size, &amp;frev,
							   &amp;crev, &amp;igp_offset)) {

					if (crev &gt;= 2) {
						igp_obj =
							(ATOM_INTEGRATED_SYSTEM_INFO_V2
							 *) (ctx-&gt;bios + igp_offset);

						if (igp_obj) {
							uint32_t slot_config, ct;

							if (con_obj_num == 1)
								slot_config =
									igp_obj-&gt;
									ulDDISlot1Config;
							else
								slot_config =
									igp_obj-&gt;
									ulDDISlot2Config;

							ct = (slot_config &gt;&gt; 16) &amp; 0xff;
							connector_type =
								object_connector_convert
								[ct];
							connector_object_id = ct;
							igp_lane_info =
								slot_config &amp; 0xffff;
						} else
							continue;
					} else
						continue;
				} else {
					igp_lane_info = 0;
					connector_type =
						object_connector_convert[con_obj_id];
					connector_object_id = con_obj_id;
				}
			} else {
				igp_lane_info = 0;
				connector_type =
				    object_connector_convert[con_obj_id];
				connector_object_id = con_obj_id;
			}

			if (connector_type == DRM_MODE_CONNECTOR_Unknown)
				continue;

			for (j = 0; j &lt; ((le16_to_cpu(path-&gt;usSize) - 8) / 2);
			     j++) {
				uint8_t enc_obj_id, enc_obj_num, enc_obj_type;

				enc_obj_id =
				    (le16_to_cpu(path-&gt;usGraphicObjIds[j]) &amp;
				     OBJECT_ID_MASK) &gt;&gt; OBJECT_ID_SHIFT;
				enc_obj_num =
				    (le16_to_cpu(path-&gt;usGraphicObjIds[j]) &amp;
				     ENUM_ID_MASK) &gt;&gt; ENUM_ID_SHIFT;
				enc_obj_type =
				    (le16_to_cpu(path-&gt;usGraphicObjIds[j]) &amp;
				     OBJECT_TYPE_MASK) &gt;&gt; OBJECT_TYPE_SHIFT;

				/* FIXME: add support for router objects */
				if (enc_obj_type == GRAPH_OBJECT_TYPE_ENCODER) {
					if (enc_obj_num == 2)
						linkb = true;
					else
						linkb = false;

					radeon_add_atom_encoder(dev,
								enc_obj_id,
								le16_to_cpu
								(path-&gt;
								 usDeviceTag));

				}
			}

			/* look up gpio for ddc, hpd */
			if ((le16_to_cpu(path-&gt;usDeviceTag) &amp;
			     (ATOM_DEVICE_TV_SUPPORT | ATOM_DEVICE_CV_SUPPORT)) == 0) {
				for (j = 0; j &lt; con_obj-&gt;ucNumberOfObjects; j++) {
					if (le16_to_cpu(path-&gt;usConnObjectId) ==
					    le16_to_cpu(con_obj-&gt;asObjects[j].
							usObjectID)) {
						ATOM_COMMON_RECORD_HEADER
						    *record =
						    (ATOM_COMMON_RECORD_HEADER
						     *)
						    (ctx-&gt;bios + data_offset +
						     le16_to_cpu(con_obj-&gt;
								 asObjects[j].
								 usRecordOffset));
						ATOM_I2C_RECORD *i2c_record;
						ATOM_HPD_INT_RECORD *hpd_record;
						ATOM_I2C_ID_CONFIG_ACCESS *i2c_config;
						hpd.hpd = RADEON_HPD_NONE;

						while (record-&gt;ucRecordType &gt; 0
						       &amp;&amp; record-&gt;
						       ucRecordType &lt;=
						       ATOM_MAX_OBJECT_RECORD_NUMBER) {
							switch (record-&gt;ucRecordType) {
							case ATOM_I2C_RECORD_TYPE:
								i2c_record =
								    (ATOM_I2C_RECORD *)
									record;
								i2c_config =
									(ATOM_I2C_ID_CONFIG_ACCESS *)
									&amp;i2c_record-&gt;sucI2cId;
								ddc_bus = radeon_lookup_i2c_gpio(rdev,
												 i2c_config-&gt;
												 ucAccess);
								break;
							case ATOM_HPD_INT_RECORD_TYPE:
								hpd_record =
									(ATOM_HPD_INT_RECORD *)
									record;
								gpio = radeon_lookup_gpio(rdev,
											  hpd_record-&gt;ucHPDIntGPIOID);
								hpd = radeon_atom_get_hpd_info_from_gpio(rdev, &amp;gpio);
								hpd.plugged_state = hpd_record-&gt;ucPlugged_PinState;
								break;
							}
							record =
							    (ATOM_COMMON_RECORD_HEADER
							     *) ((char *)record
								 +
								 record-&gt;
								 ucRecordSize);
						}
						break;
					}
				}
			} else {
				hpd.hpd = RADEON_HPD_NONE;
				ddc_bus.valid = false;
			}

			/* needed for aux chan transactions */
			ddc_bus.hpd_id = hpd.hpd ? (hpd.hpd - 1) : 0;

			conn_id = le16_to_cpu(path-&gt;usConnObjectId);

			if (!radeon_atom_apply_quirks
			    (dev, le16_to_cpu(path-&gt;usDeviceTag), &amp;connector_type,
			     &amp;ddc_bus, &amp;conn_id, &amp;hpd))
				continue;

			radeon_add_atom_connector(dev,
						  conn_id,
						  le16_to_cpu(path-&gt;
							      usDeviceTag),
						  connector_type, &amp;ddc_bus,
						  linkb, igp_lane_info,
						  connector_object_id,
						  &amp;hpd);

		}
	}

	radeon_link_encoder_connector(dev);

	return true;
}

static uint16_t atombios_get_connector_object_id(struct drm_device *dev,
						 int connector_type,
						 uint16_t devices)
{
	struct radeon_device *rdev = dev-&gt;dev_private;

	if (rdev-&gt;flags &amp; RADEON_IS_IGP) {
		return supported_devices_connector_object_id_convert
			[connector_type];
	} else if (((connector_type == DRM_MODE_CONNECTOR_DVII) ||
		    (connector_type == DRM_MODE_CONNECTOR_DVID)) &amp;&amp;
		   (devices &amp; ATOM_DEVICE_DFP2_SUPPORT))  {
		struct radeon_mode_info *mode_info = &amp;rdev-&gt;mode_info;
		struct atom_context *ctx = mode_info-&gt;atom_context;
		int index = GetIndexIntoMasterTable(DATA, XTMDS_Info);
		uint16_t size, data_offset;
		uint8_t frev, crev;
		ATOM_XTMDS_INFO *xtmds;

		if (atom_parse_data_header(ctx, index, &amp;size, &amp;frev, &amp;crev, &amp;data_offset)) {
			xtmds = (ATOM_XTMDS_INFO *)(ctx-&gt;bios + data_offset);

			if (xtmds-&gt;ucSupportedLink &amp; ATOM_XTMDS_SUPPORTED_DUALLINK) {
				if (connector_type == DRM_MODE_CONNECTOR_DVII)
					return CONNECTOR_OBJECT_ID_DUAL_LINK_DVI_I;
				else
					return CONNECTOR_OBJECT_ID_DUAL_LINK_DVI_D;
			} else {
				if (connector_type == DRM_MODE_CONNECTOR_DVII)
					return CONNECTOR_OBJECT_ID_SINGLE_LINK_DVI_I;
				else
					return CONNECTOR_OBJECT_ID_SINGLE_LINK_DVI_D;
			}
		} else
			return supported_devices_connector_object_id_convert
				[connector_type];
	} else {
		return supported_devices_connector_object_id_convert
			[connector_type];
	}
}

struct bios_connector {
	bool valid;
	uint16_t line_mux;
	uint16_t devices;
	int connector_type;
	struct radeon_i2c_bus_rec ddc_bus;
	struct radeon_hpd hpd;
};

bool radeon_get_atom_connector_info_from_supported_devices_table(struct
								 drm_device
								 *dev)
{
	struct radeon_device *rdev = dev-&gt;dev_private;
	struct radeon_mode_info *mode_info = &amp;rdev-&gt;mode_info;
	struct atom_context *ctx = mode_info-&gt;atom_context;
	int index = GetIndexIntoMasterTable(DATA, SupportedDevicesInfo);
	uint16_t size, data_offset;
	uint8_t frev, crev;
	uint16_t device_support;
	uint8_t dac;
	union atom_supported_devices *supported_devices;
	int i, j, max_device;
	struct bios_connector bios_connectors[ATOM_MAX_SUPPORTED_DEVICE];

	if (!atom_parse_data_header(ctx, index, &amp;size, &amp;frev, &amp;crev, &amp;data_offset))
		return false;

	supported_devices =
	    (union atom_supported_devices *)(ctx-&gt;bios + data_offset);

	device_support = le16_to_cpu(supported_devices-&gt;info.usDeviceSupport);

	if (frev &gt; 1)
		max_device = ATOM_MAX_SUPPORTED_DEVICE;
	else
		max_device = ATOM_MAX_SUPPORTED_DEVICE_INFO;

	for (i = 0; i &lt; max_device; i++) {
		ATOM_CONNECTOR_INFO_I2C ci =
		    supported_devices-&gt;info.asConnInfo[i];

		bios_connectors[i].valid = false;

		if (!(device_support &amp; (1 &lt;&lt; i))) {
			continue;
		}

		if (i == ATOM_DEVICE_CV_INDEX) {
			DRM_DEBUG("Skipping Component Video\n");
			continue;
		}

		bios_connectors[i].connector_type =
		    supported_devices_connector_convert[ci.sucConnectorInfo.
							sbfAccess.
							bfConnectorType];

		if (bios_connectors[i].connector_type ==
		    DRM_MODE_CONNECTOR_Unknown)
			continue;

		dac = ci.sucConnectorInfo.sbfAccess.bfAssociatedDAC;

		bios_connectors[i].line_mux =
			ci.sucI2cId.ucAccess;

		/* give tv unique connector ids */
		if (i == ATOM_DEVICE_TV1_INDEX) {
			bios_connectors[i].ddc_bus.valid = false;
			bios_connectors[i].line_mux = 50;
		} else if (i == ATOM_DEVICE_TV2_INDEX) {
			bios_connectors[i].ddc_bus.valid = false;
			bios_connectors[i].line_mux = 51;
		} else if (i == ATOM_DEVICE_CV_INDEX) {
			bios_connectors[i].ddc_bus.valid = false;
			bios_connectors[i].line_mux = 52;
		} else
			bios_connectors[i].ddc_bus =
			    radeon_lookup_i2c_gpio(rdev,
						   bios_connectors[i].line_mux);

		if ((crev &gt; 1) &amp;&amp; (frev &gt; 1)) {
			u8 isb = supported_devices-&gt;info_2d1.asIntSrcInfo[i].ucIntSrcBitmap;
			switch (isb) {
			case 0x4:
				bios_connectors[i].hpd.hpd = RADEON_HPD_1;
				break;
			case 0xa:
				bios_connectors[i].hpd.hpd = RADEON_HPD_2;
				break;
			default:
				bios_connectors[i].hpd.hpd = RADEON_HPD_NONE;
				break;
			}
		} else {
			if (i == ATOM_DEVICE_DFP1_INDEX)
				bios_connectors[i].hpd.hpd = RADEON_HPD_1;
			else if (i == ATOM_DEVICE_DFP2_INDEX)
				bios_connectors[i].hpd.hpd = RADEON_HPD_2;
			else
				bios_connectors[i].hpd.hpd = RADEON_HPD_NONE;
		}

		/* Always set the connector type to VGA for CRT1/CRT2. if they are
		 * shared with a DVI port, we'll pick up the DVI connector when we
		 * merge the outputs.  Some bioses incorrectly list VGA ports as DVI.
		 */
		if (i == ATOM_DEVICE_CRT1_INDEX || i == ATOM_DEVICE_CRT2_INDEX)
			bios_connectors[i].connector_type =
			    DRM_MODE_CONNECTOR_VGA;

		if (!radeon_atom_apply_quirks
		    (dev, (1 &lt;&lt; i), &amp;bios_connectors[i].connector_type,
		     &amp;bios_connectors[i].ddc_bus, &amp;bios_connectors[i].line_mux,
		     &amp;bios_connectors[i].hpd))
			continue;

		bios_connectors[i].valid = true;
		bios_connectors[i].devices = (1 &lt;&lt; i);

		if (ASIC_IS_AVIVO(rdev) || radeon_r4xx_atom)
			radeon_add_atom_encoder(dev,
						radeon_get_encoder_id(dev,
								      (1 &lt;&lt; i),
								      dac),
						(1 &lt;&lt; i));
		else
			radeon_add_legacy_encoder(dev,
						  radeon_get_encoder_id(dev,
									(1 &lt;&lt; i),
									dac),
						  (1 &lt;&lt; i));
	}

	/* combine shared connectors */
	for (i = 0; i &lt; max_device; i++) {
		if (bios_connectors[i].valid) {
			for (j = 0; j &lt; max_device; j++) {
				if (bios_connectors[j].valid &amp;&amp; (i != j)) {
					if (bios_connectors[i].line_mux ==
					    bios_connectors[j].line_mux) {
						/* make sure not to combine LVDS */
						if (bios_connectors[i].devices &amp; (ATOM_DEVICE_LCD_SUPPORT)) {
							bios_connectors[i].line_mux = 53;
							bios_connectors[i].ddc_bus.valid = false;
							continue;
						}
						if (bios_connectors[j].devices &amp; (ATOM_DEVICE_LCD_SUPPORT)) {
							bios_connectors[j].line_mux = 53;
							bios_connectors[j].ddc_bus.valid = false;
							continue;
						}
						/* combine analog and digital for DVI-I */
						if (((bios_connectors[i].devices &amp; (ATOM_DEVICE_DFP_SUPPORT)) &amp;&amp;
						     (bios_connectors[j].devices &amp; (ATOM_DEVICE_CRT_SUPPORT))) ||
						    ((bios_connectors[j].devices &amp; (ATOM_DEVICE_DFP_SUPPORT)) &amp;&amp;
						     (bios_connectors[i].devices &amp; (ATOM_DEVICE_CRT_SUPPORT)))) {
							bios_connectors[i].devices |=
								bios_connectors[j].devices;
							bios_connectors[i].connector_type =
								DRM_MODE_CONNECTOR_DVII;
							if (bios_connectors[j].devices &amp; (ATOM_DEVICE_DFP_SUPPORT))
								bios_connectors[i].hpd =
									bios_connectors[j].hpd;
							bios_connectors[j].valid = false;
						}
					}
				}
			}
		}
	}

	/* add the connectors */
	for (i = 0; i &lt; max_device; i++) {
		if (bios_connectors[i].valid) {
			uint16_t connector_object_id =
				atombios_get_connector_object_id(dev,
						      bios_connectors[i].connector_type,
						      bios_connectors[i].devices);
			radeon_add_atom_connector(dev,
						  bios_connectors[i].line_mux,
						  bios_connectors[i].devices,
						  bios_connectors[i].
						  connector_type,
						  &amp;bios_connectors[i].ddc_bus,
						  false, 0,
						  connector_object_id,
						  &amp;bios_connectors[i].hpd);
		}
	}

	radeon_link_encoder_connector(dev);

	return true;
}

union firmware_info {
	ATOM_FIRMWARE_INFO info;
	ATOM_FIRMWARE_INFO_V1_2 info_12;
	ATOM_FIRMWARE_INFO_V1_3 info_13;
	ATOM_FIRMWARE_INFO_V1_4 info_14;
	ATOM_FIRMWARE_INFO_V2_1 info_21;
};

bool radeon_atom_get_clock_info(struct drm_device *dev)
{
	struct radeon_device *rdev = dev-&gt;dev_private;
	struct radeon_mode_info *mode_info = &amp;rdev-&gt;mode_info;
	int index = GetIndexIntoMasterTable(DATA, FirmwareInfo);
	union firmware_info *firmware_info;
	uint8_t frev, crev;
	struct radeon_pll *p1pll = &amp;rdev-&gt;clock.p1pll;
	struct radeon_pll *p2pll = &amp;rdev-&gt;clock.p2pll;
	struct radeon_pll *dcpll = &amp;rdev-&gt;clock.dcpll;
	struct radeon_pll *spll = &amp;rdev-&gt;clock.spll;
	struct radeon_pll *mpll = &amp;rdev-&gt;clock.mpll;
	uint16_t data_offset;

	if (atom_parse_data_header(mode_info-&gt;atom_context, index, NULL,
				   &amp;frev, &amp;crev, &amp;data_offset)) {
		firmware_info =
			(union firmware_info *)(mode_info-&gt;atom_context-&gt;bios +
						data_offset);
		/* pixel clocks */
		p1pll-&gt;reference_freq =
		    le16_to_cpu(firmware_info-&gt;info.usReferenceClock);
		p1pll-&gt;reference_div = 0;

		if (crev &lt; 2)
			p1pll-&gt;pll_out_min =
				le16_to_cpu(firmware_info-&gt;info.usMinPixelClockPLL_Output);
		else
			p1pll-&gt;pll_out_min =
				le32_to_cpu(firmware_info-&gt;info_12.ulMinPixelClockPLL_Output);
		p1pll-&gt;pll_out_max =
		    le32_to_cpu(firmware_info-&gt;info.ulMaxPixelClockPLL_Output);

		if (crev &gt;= 4) {
			p1pll-&gt;lcd_pll_out_min =
				le16_to_cpu(firmware_info-&gt;info_14.usLcdMinPixelClockPLL_Output) * 100;
			if (p1pll-&gt;lcd_pll_out_min == 0)
				p1pll-&gt;lcd_pll_out_min = p1pll-&gt;pll_out_min;
			p1pll-&gt;lcd_pll_out_max =
				le16_to_cpu(firmware_info-&gt;info_14.usLcdMaxPixelClockPLL_Output) * 100;
			if (p1pll-&gt;lcd_pll_out_max == 0)
				p1pll-&gt;lcd_pll_out_max = p1pll-&gt;pll_out_max;
		} else {
			p1pll-&gt;lcd_pll_out_min = p1pll-&gt;pll_out_min;
			p1pll-&gt;lcd_pll_out_max = p1pll-&gt;pll_out_max;
		}

		if (p1pll-&gt;pll_out_min == 0) {
			if (ASIC_IS_AVIVO(rdev))
				p1pll-&gt;pll_out_min = 64800;
			else
				p1pll-&gt;pll_out_min = 20000;
		} else if (p1pll-&gt;pll_out_min &gt; 64800) {
			/* Limiting the pll output range is a good thing generally as
			 * it limits the number of possible pll combinations for a given
			 * frequency presumably to the ones that work best on each card.
			 * However, certain duallink DVI monitors seem to like
			 * pll combinations that would be limited by this at least on
			 * pre-DCE 3.0 r6xx hardware.  This might need to be adjusted per
			 * family.
			 */
			if (!radeon_new_pll)
				p1pll-&gt;pll_out_min = 64800;
		}

		p1pll-&gt;pll_in_min =
		    le16_to_cpu(firmware_info-&gt;info.usMinPixelClockPLL_Input);
		p1pll-&gt;pll_in_max =
		    le16_to_cpu(firmware_info-&gt;info.usMaxPixelClockPLL_Input);

		*p2pll = *p1pll;

		/* system clock */
		spll-&gt;reference_freq =
		    le16_to_cpu(firmware_info-&gt;info.usReferenceClock);
		spll-&gt;reference_div = 0;

		spll-&gt;pll_out_min =
		    le16_to_cpu(firmware_info-&gt;info.usMinEngineClockPLL_Output);
		spll-&gt;pll_out_max =
		    le32_to_cpu(firmware_info-&gt;info.ulMaxEngineClockPLL_Output);

		/* ??? */
		if (spll-&gt;pll_out_min == 0) {
			if (ASIC_IS_AVIVO(rdev))
				spll-&gt;pll_out_min = 64800;
			else
				spll-&gt;pll_out_min = 20000;
		}

		spll-&gt;pll_in_min =
		    le16_to_cpu(firmware_info-&gt;info.usMinEngineClockPLL_Input);
		spll-&gt;pll_in_max =
		    le16_to_cpu(firmware_info-&gt;info.usMaxEngineClockPLL_Input);

		/* memory clock */
		mpll-&gt;reference_freq =
		    le16_to_cpu(firmware_info-&gt;info.usReferenceClock);
		mpll-&gt;reference_div = 0;

		mpll-&gt;pll_out_min =
		    le16_to_cpu(firmware_info-&gt;info.usMinMemoryClockPLL_Output);
		mpll-&gt;pll_out_max =
		    le32_to_cpu(firmware_info-&gt;info.ulMaxMemoryClockPLL_Output);

		/* ??? */
		if (mpll-&gt;pll_out_min == 0) {
			if (ASIC_IS_AVIVO(rdev))
				mpll-&gt;pll_out_min = 64800;
			else
				mpll-&gt;pll_out_min = 20000;
		}

		mpll-&gt;pll_in_min =
		    le16_to_cpu(firmware_info-&gt;info.usMinMemoryClockPLL_Input);
		mpll-&gt;pll_in_max =
		    le16_to_cpu(firmware_info-&gt;info.usMaxMemoryClockPLL_Input);

		rdev-&gt;clock.default_sclk =
		    le32_to_cpu(firmware_info-&gt;info.ulDefaultEngineClock);
		rdev-&gt;clock.default_mclk =
		    le32_to_cpu(firmware_info-&gt;info.ulDefaultMemoryClock);

		if (ASIC_IS_DCE4(rdev)) {
			rdev-&gt;clock.default_dispclk =
				le32_to_cpu(firmware_info-&gt;info_21.ulDefaultDispEngineClkFreq);
			if (rdev-&gt;clock.default_dispclk == 0)
				rdev-&gt;clock.default_dispclk = 60000; /* 600 Mhz */
			rdev-&gt;clock.dp_extclk =
				le16_to_cpu(firmware_info-&gt;info_21.usUniphyDPModeExtClkFreq);
		}
		*dcpll = *p1pll;

		return true;
	}

	return false;
}

union igp_info {
	struct _ATOM_INTEGRATED_SYSTEM_INFO info;
	struct _ATOM_INTEGRATED_SYSTEM_INFO_V2 info_2;
};

bool radeon_atombios_sideport_present(struct radeon_device *rdev)
{
	struct radeon_mode_info *mode_info = &amp;rdev-&gt;mode_info;
	int index = GetIndexIntoMasterTable(DATA, IntegratedSystemInfo);
	union igp_info *igp_info;
	u8 frev, crev;
	u16 data_offset;

	if (atom_parse_data_header(mode_info-&gt;atom_context, index, NULL,
				   &amp;frev, &amp;crev, &amp;data_offset)) {
		igp_info = (union igp_info *)(mode_info-&gt;atom_context-&gt;bios +
				      data_offset);
		switch (crev) {
		case 1:
			if (igp_info-&gt;info.ucMemoryType &amp; 0xf0)
				return true;
			break;
		case 2:
			if (igp_info-&gt;info_2.ucMemoryType &amp; 0x0f)
				return true;
			break;
		default:
			DRM_ERROR("Unsupported IGP table: %d %d\n", frev, crev);
			break;
		}
	}
	return false;
}

bool radeon_atombios_get_tmds_info(struct radeon_encoder *encoder,
				   struct radeon_encoder_int_tmds *tmds)
{
	struct drm_device *dev = encoder-&gt;base.dev;
	struct radeon_device *rdev = dev-&gt;dev_private;
	struct radeon_mode_info *mode_info = &amp;rdev-&gt;mode_info;
	int index = GetIndexIntoMasterTable(DATA, TMDS_Info);
	uint16_t data_offset;
	struct _ATOM_TMDS_INFO *tmds_info;
	uint8_t frev, crev;
	uint16_t maxfreq;
	int i;

	if (atom_parse_data_header(mode_info-&gt;atom_context, index, NULL,
				   &amp;frev, &amp;crev, &amp;data_offset)) {
		tmds_info =
			(struct _ATOM_TMDS_INFO *)(mode_info-&gt;atom_context-&gt;bios +
						   data_offset);

		maxfreq = le16_to_cpu(tmds_info-&gt;usMaxFrequency);
		for (i = 0; i &lt; 4; i++) {
			tmds-&gt;tmds_pll[i].freq =
			    le16_to_cpu(tmds_info-&gt;asMiscInfo[i].usFrequency);
			tmds-&gt;tmds_pll[i].value =
			    tmds_info-&gt;asMiscInfo[i].ucPLL_ChargePump &amp; 0x3f;
			tmds-&gt;tmds_pll[i].value |=
			    (tmds_info-&gt;asMiscInfo[i].
			     ucPLL_VCO_Gain &amp; 0x3f) &lt;&lt; 6;
			tmds-&gt;tmds_pll[i].value |=
			    (tmds_info-&gt;asMiscInfo[i].
			     ucPLL_DutyCycle &amp; 0xf) &lt;&lt; 12;
			tmds-&gt;tmds_pll[i].value |=
			    (tmds_info-&gt;asMiscInfo[i].
			     ucPLL_VoltageSwing &amp; 0xf) &lt;&lt; 16;

			DRM_DEBUG("TMDS PLL From ATOMBIOS %u %x\n",
				  tmds-&gt;tmds_pll[i].freq,
				  tmds-&gt;tmds_pll[i].value);

			if (maxfreq == tmds-&gt;tmds_pll[i].freq) {
				tmds-&gt;tmds_pll[i].freq = 0xffffffff;
				break;
			}
		}
		return true;
	}
	return false;
}

static struct radeon_atom_ss *radeon_atombios_get_ss_info(struct
							  radeon_encoder
							  *encoder,
							  int id)
{
	struct drm_device *dev = encoder-&gt;base.dev;
	struct radeon_device *rdev = dev-&gt;dev_private;
	struct radeon_mode_info *mode_info = &amp;rdev-&gt;mode_info;
	int index = GetIndexIntoMasterTable(DATA, PPLL_SS_Info);
	uint16_t data_offset;
	struct _ATOM_SPREAD_SPECTRUM_INFO *ss_info;
	uint8_t frev, crev;
	struct radeon_atom_ss *ss = NULL;
	int i;

	if (id &gt; ATOM_MAX_SS_ENTRY)
		return NULL;

	if (atom_parse_data_header(mode_info-&gt;atom_context, index, NULL,
				   &amp;frev, &amp;crev, &amp;data_offset)) {
		ss_info =
			(struct _ATOM_SPREAD_SPECTRUM_INFO *)(mode_info-&gt;atom_context-&gt;bios + data_offset);

		ss =
		    kzalloc(sizeof(struct radeon_atom_ss), GFP_KERNEL);

		if (!ss)
			return NULL;

		for (i = 0; i &lt; ATOM_MAX_SS_ENTRY; i++) {
			if (ss_info-&gt;asSS_Info[i].ucSS_Id == id) {
				ss-&gt;percentage =
					le16_to_cpu(ss_info-&gt;asSS_Info[i].usSpreadSpectrumPercentage);
				ss-&gt;type = ss_info-&gt;asSS_Info[i].ucSpreadSpectrumType;
				ss-&gt;step = ss_info-&gt;asSS_Info[i].ucSS_Step;
				ss-&gt;delay = ss_info-&gt;asSS_Info[i].ucSS_Delay;
				ss-&gt;range = ss_info-&gt;asSS_Info[i].ucSS_Range;
				ss-&gt;refdiv = ss_info-&gt;asSS_Info[i].ucRecommendedRef_Div;
				break;
			}
		}
	}
	return ss;
}

union lvds_info {
	struct _ATOM_LVDS_INFO info;
	struct _ATOM_LVDS_INFO_V12 info_12;
};

struct radeon_encoder_atom_dig *radeon_atombios_get_lvds_info(struct
							      radeon_encoder
							      *encoder)
{
	struct drm_device *dev = encoder-&gt;base.dev;
	struct radeon_device *rdev = dev-&gt;dev_private;
	struct radeon_mode_info *mode_info = &amp;rdev-&gt;mode_info;
	int index = GetIndexIntoMasterTable(DATA, LVDS_Info);
	uint16_t data_offset, misc;
	union lvds_info *lvds_info;
	uint8_t frev, crev;
	struct radeon_encoder_atom_dig *lvds = NULL;

	if (atom_parse_data_header(mode_info-&gt;atom_context, index, NULL,
				   &amp;frev, &amp;crev, &amp;data_offset)) {
		lvds_info =
			(union lvds_info *)(mode_info-&gt;atom_context-&gt;bios + data_offset);
		lvds =
		    kzalloc(sizeof(struct radeon_encoder_atom_dig), GFP_KERNEL);

		if (!lvds)
			return NULL;

		lvds-&gt;native_mode.clock =
		    le16_to_cpu(lvds_info-&gt;info.sLCDTiming.usPixClk) * 10;
		lvds-&gt;native_mode.hdisplay =
		    le16_to_cpu(lvds_info-&gt;info.sLCDTiming.usHActive);
		lvds-&gt;native_mode.vdisplay =
		    le16_to_cpu(lvds_info-&gt;info.sLCDTiming.usVActive);
		lvds-&gt;native_mode.htotal = lvds-&gt;native_mode.hdisplay +
			le16_to_cpu(lvds_info-&gt;info.sLCDTiming.usHBlanking_Time);
		lvds-&gt;native_mode.hsync_start = lvds-&gt;native_mode.hdisplay +
			le16_to_cpu(lvds_info-&gt;info.sLCDTiming.usHSyncOffset);
		lvds-&gt;native_mode.hsync_end = lvds-&gt;native_mode.hsync_start +
			le16_to_cpu(lvds_info-&gt;info.sLCDTiming.usHSyncWidth);
		lvds-&gt;native_mode.vtotal = lvds-&gt;native_mode.vdisplay +
			le16_to_cpu(lvds_info-&gt;info.sLCDTiming.usVBlanking_Time);
		lvds-&gt;native_mode.vsync_start = lvds-&gt;native_mode.vdisplay +
			le16_to_cpu(lvds_info-&gt;info.sLCDTiming.usVSyncWidth);
		lvds-&gt;native_mode.vsync_end = lvds-&gt;native_mode.vsync_start +
			le16_to_cpu(lvds_info-&gt;info.sLCDTiming.usVSyncWidth);
		lvds-&gt;panel_pwr_delay =
		    le16_to_cpu(lvds_info-&gt;info.usOffDelayInMs);
		lvds-&gt;lvds_misc = lvds_info-&gt;info.ucLVDS_Misc;

		misc = le16_to_cpu(lvds_info-&gt;info.sLCDTiming.susModeMiscInfo.usAccess);
		if (misc &amp; ATOM_VSYNC_POLARITY)
			lvds-&gt;native_mode.flags |= DRM_MODE_FLAG_NVSYNC;
		if (misc &amp; ATOM_HSYNC_POLARITY)
			lvds-&gt;native_mode.flags |= DRM_MODE_FLAG_NHSYNC;
		if (misc &amp; ATOM_COMPOSITESYNC)
			lvds-&gt;native_mode.flags |= DRM_MODE_FLAG_CSYNC;
		if (misc &amp; ATOM_INTERLACE)
			lvds-&gt;native_mode.flags |= DRM_MODE_FLAG_INTERLACE;
		if (misc &amp; ATOM_DOUBLE_CLOCK_MODE)
			lvds-&gt;native_mode.flags |= DRM_MODE_FLAG_DBLSCAN;

		/* set crtc values */
		drm_mode_set_crtcinfo(&amp;lvds-&gt;native_mode, CRTC_INTERLACE_HALVE_V);

		lvds-&gt;ss = radeon_atombios_get_ss_info(encoder, lvds_info-&gt;info.ucSS_Id);

		if (ASIC_IS_AVIVO(rdev)) {
			if (radeon_new_pll == 0)
				lvds-&gt;pll_algo = PLL_ALGO_LEGACY;
			else
				lvds-&gt;pll_algo = PLL_ALGO_NEW;
		} else {
			if (radeon_new_pll == 1)
				lvds-&gt;pll_algo = PLL_ALGO_NEW;
			else
				lvds-&gt;pll_algo = PLL_ALGO_LEGACY;
		}

		encoder-&gt;native_mode = lvds-&gt;native_mode;
	}
	return lvds;
}

struct radeon_encoder_primary_dac *
radeon_atombios_get_primary_dac_info(struct radeon_encoder *encoder)
{
	struct drm_device *dev = encoder-&gt;base.dev;
	struct radeon_device *rdev = dev-&gt;dev_private;
	struct radeon_mode_info *mode_info = &amp;rdev-&gt;mode_info;
	int index = GetIndexIntoMasterTable(DATA, CompassionateData);
	uint16_t data_offset;
	struct _COMPASSIONATE_DATA *dac_info;
	uint8_t frev, crev;
	uint8_t bg, dac;
	struct radeon_encoder_primary_dac *p_dac = NULL;

	if (atom_parse_data_header(mode_info-&gt;atom_context, index, NULL,
				   &amp;frev, &amp;crev, &amp;data_offset)) {
		dac_info = (struct _COMPASSIONATE_DATA *)
			(mode_info-&gt;atom_context-&gt;bios + data_offset);

		p_dac = kzalloc(sizeof(struct radeon_encoder_primary_dac), GFP_KERNEL);

		if (!p_dac)
			return NULL;

		bg = dac_info-&gt;ucDAC1_BG_Adjustment;
		dac = dac_info-&gt;ucDAC1_DAC_Adjustment;
		p_dac-&gt;ps2_pdac_adj = (bg &lt;&lt; 8) | (dac);

	}
	return p_dac;
}

bool radeon_atom_get_tv_timings(struct radeon_device *rdev, int index,
				struct drm_display_mode *mode)
{
	struct radeon_mode_info *mode_info = &amp;rdev-&gt;mode_info;
	ATOM_ANALOG_TV_INFO *tv_info;
	ATOM_ANALOG_TV_INFO_V1_2 *tv_info_v1_2;
	ATOM_DTD_FORMAT *dtd_timings;
	int data_index = GetIndexIntoMasterTable(DATA, AnalogTV_Info);
	u8 frev, crev;
	u16 data_offset, misc;

	if (!atom_parse_data_header(mode_info-&gt;atom_context, data_index, NULL,
				    &amp;frev, &amp;crev, &amp;data_offset))
		return false;

	switch (crev) {
	case 1:
		tv_info = (ATOM_ANALOG_TV_INFO *)(mode_info-&gt;atom_context-&gt;bios + data_offset);
		if (index &gt;= MAX_SUPPORTED_TV_TIMING)
			return false;

		mode-&gt;crtc_htotal = le16_to_cpu(tv_info-&gt;aModeTimings[index].usCRTC_H_Total);
		mode-&gt;crtc_hdisplay = le16_to_cpu(tv_info-&gt;aModeTimings[index].usCRTC_H_Disp);
		mode-&gt;crtc_hsync_start = le16_to_cpu(tv_info-&gt;aModeTimings[index].usCRTC_H_SyncStart);
		mode-&gt;crtc_hsync_end = le16_to_cpu(tv_info-&gt;aModeTimings[index].usCRTC_H_SyncStart) +
			le16_to_cpu(tv_info-&gt;aModeTimings[index].usCRTC_H_SyncWidth);

		mode-&gt;crtc_vtotal = le16_to_cpu(tv_info-&gt;aModeTimings[index].usCRTC_V_Total);
		mode-&gt;crtc_vdisplay = le16_to_cpu(tv_info-&gt;aModeTimings[index].usCRTC_V_Disp);
		mode-&gt;crtc_vsync_start = le16_to_cpu(tv_info-&gt;aModeTimings[index].usCRTC_V_SyncStart);
		mode-&gt;crtc_vsync_end = le16_to_cpu(tv_info-&gt;aModeTimings[index].usCRTC_V_SyncStart) +
			le16_to_cpu(tv_info-&gt;aModeTimings[index].usCRTC_V_SyncWidth);

		mode-&gt;flags = 0;
		misc = le16_to_cpu(tv_info-&gt;aModeTimings[index].susModeMiscInfo.usAccess);
		if (misc &amp; ATOM_VSYNC_POLARITY)
			mode-&gt;flags |= DRM_MODE_FLAG_NVSYNC;
		if (misc &amp; ATOM_HSYNC_POLARITY)
			mode-&gt;flags |= DRM_MODE_FLAG_NHSYNC;
		if (misc &amp; ATOM_COMPOSITESYNC)
			mode-&gt;flags |= DRM_MODE_FLAG_CSYNC;
		if (misc &amp; ATOM_INTERLACE)
			mode-&gt;flags |= DRM_MODE_FLAG_INTERLACE;
		if (misc &amp; ATOM_DOUBLE_CLOCK_MODE)
			mode-&gt;flags |= DRM_MODE_FLAG_DBLSCAN;

		mode-&gt;clock = le16_to_cpu(tv_info-&gt;aModeTimings[index].usPixelClock) * 10;

		if (index == 1) {
			/* PAL timings appear to have wrong values for totals */
			mode-&gt;crtc_htotal -= 1;
			mode-&gt;crtc_vtotal -= 1;
		}
		break;
	case 2:
		tv_info_v1_2 = (ATOM_ANALOG_TV_INFO_V1_2 *)(mode_info-&gt;atom_context-&gt;bios + data_offset);
		if (index &gt;= MAX_SUPPORTED_TV_TIMING_V1_2)
			return false;

		dtd_timings = &amp;tv_info_v1_2-&gt;aModeTimings[index];
		mode-&gt;crtc_htotal = le16_to_cpu(dtd_timings-&gt;usHActive) +
			le16_to_cpu(dtd_timings-&gt;usHBlanking_Time);
		mode-&gt;crtc_hdisplay = le16_to_cpu(dtd_timings-&gt;usHActive);
		mode-&gt;crtc_hsync_start = le16_to_cpu(dtd_timings-&gt;usHActive) +
			le16_to_cpu(dtd_timings-&gt;usHSyncOffset);
		mode-&gt;crtc_hsync_end = mode-&gt;crtc_hsync_start +
			le16_to_cpu(dtd_timings-&gt;usHSyncWidth);

		mode-&gt;crtc_vtotal = le16_to_cpu(dtd_timings-&gt;usVActive) +
			le16_to_cpu(dtd_timings-&gt;usVBlanking_Time);
		mode-&gt;crtc_vdisplay = le16_to_cpu(dtd_timings-&gt;usVActive);
		mode-&gt;crtc_vsync_start = le16_to_cpu(dtd_timings-&gt;usVActive) +
			le16_to_cpu(dtd_timings-&gt;usVSyncOffset);
		mode-&gt;crtc_vsync_end = mode-&gt;crtc_vsync_start +
			le16_to_cpu(dtd_timings-&gt;usVSyncWidth);

		mode-&gt;flags = 0;
		misc = le16_to_cpu(dtd_timings-&gt;susModeMiscInfo.usAccess);
		if (misc &amp; ATOM_VSYNC_POLARITY)
			mode-&gt;flags |= DRM_MODE_FLAG_NVSYNC;
		if (misc &amp; ATOM_HSYNC_POLARITY)
			mode-&gt;flags |= DRM_MODE_FLAG_NHSYNC;
		if (misc &amp; ATOM_COMPOSITESYNC)
			mode-&gt;flags |= DRM_MODE_FLAG_CSYNC;
		if (misc &amp; ATOM_INTERLACE)
			mode-&gt;flags |= DRM_MODE_FLAG_INTERLACE;
		if (misc &amp; ATOM_DOUBLE_CLOCK_MODE)
			mode-&gt;flags |= DRM_MODE_FLAG_DBLSCAN;

		mode-&gt;clock = le16_to_cpu(dtd_timings-&gt;usPixClk) * 10;
		break;
	}
	return true;
}

enum radeon_tv_std
radeon_atombios_get_tv_info(struct radeon_device *rdev)
{
	struct radeon_mode_info *mode_info = &amp;rdev-&gt;mode_info;
	int index = GetIndexIntoMasterTable(DATA, AnalogTV_Info);
	uint16_t data_offset;
	uint8_t frev, crev;
	struct _ATOM_ANALOG_TV_INFO *tv_info;
	enum radeon_tv_std tv_std = TV_STD_NTSC;

	if (atom_parse_data_header(mode_info-&gt;atom_context, index, NULL,
				   &amp;frev, &amp;crev, &amp;data_offset)) {

		tv_info = (struct _ATOM_ANALOG_TV_INFO *)
			(mode_info-&gt;atom_context-&gt;bios + data_offset);

		switch (tv_info-&gt;ucTV_BootUpDefaultStandard) {
		case ATOM_TV_NTSC:
			tv_std = TV_STD_NTSC;
			DRM_INFO("Default TV standard: NTSC\n");
			break;
		case ATOM_TV_NTSCJ:
			tv_std = TV_STD_NTSC_J;
			DRM_INFO("Default TV standard: NTSC-J\n");
			break;
		case ATOM_TV_PAL:
			tv_std = TV_STD_PAL;
			DRM_INFO("Default TV standard: PAL\n");
			break;
		case ATOM_TV_PALM:
			tv_std = TV_STD_PAL_M;
			DRM_INFO("Default TV standard: PAL-M\n");
			break;
		case ATOM_TV_PALN:
			tv_std = TV_STD_PAL_N;
			DRM_INFO("Default TV standard: PAL-N\n");
			break;
		case ATOM_TV_PALCN:
			tv_std = TV_STD_PAL_CN;
			DRM_INFO("Default TV standard: PAL-CN\n");
			break;
		case ATOM_TV_PAL60:
			tv_std = TV_STD_PAL_60;
			DRM_INFO("Default TV standard: PAL-60\n");
			break;
		case ATOM_TV_SECAM:
			tv_std = TV_STD_SECAM;
			DRM_INFO("Default TV standard: SECAM\n");
			break;
		default:
			tv_std = TV_STD_NTSC;
			DRM_INFO("Unknown TV standard; defaulting to NTSC\n");
			break;
		}
	}
	return tv_std;
}

struct radeon_encoder_tv_dac *
radeon_atombios_get_tv_dac_info(struct radeon_encoder *encoder)
{
	struct drm_device *dev = encoder-&gt;base.dev;
	struct radeon_device *rdev = dev-&gt;dev_private;
	struct radeon_mode_info *mode_info = &amp;rdev-&gt;mode_info;
	int index = GetIndexIntoMasterTable(DATA, CompassionateData);
	uint16_t data_offset;
	struct _COMPASSIONATE_DATA *dac_info;
	uint8_t frev, crev;
	uint8_t bg, dac;
	struct radeon_encoder_tv_dac *tv_dac = NULL;

	if (atom_parse_data_header(mode_info-&gt;atom_context, index, NULL,
				   &amp;frev, &amp;crev, &amp;data_offset)) {

		dac_info = (struct _COMPASSIONATE_DATA *)
			(mode_info-&gt;atom_context-&gt;bios + data_offset);

		tv_dac = kzalloc(sizeof(struct radeon_encoder_tv_dac), GFP_KERNEL);

		if (!tv_dac)
			return NULL;

		bg = dac_info-&gt;ucDAC2_CRT2_BG_Adjustment;
		dac = dac_info-&gt;ucDAC2_CRT2_DAC_Adjustment;
		tv_dac-&gt;ps2_tvdac_adj = (bg &lt;&lt; 16) | (dac &lt;&lt; 20);

		bg = dac_info-&gt;ucDAC2_PAL_BG_Adjustment;
		dac = dac_info-&gt;ucDAC2_PAL_DAC_Adjustment;
		tv_dac-&gt;pal_tvdac_adj = (bg &lt;&lt; 16) | (dac &lt;&lt; 20);

		bg = dac_info-&gt;ucDAC2_NTSC_BG_Adjustment;
		dac = dac_info-&gt;ucDAC2_NTSC_DAC_Adjustment;
		tv_dac-&gt;ntsc_tvdac_adj = (bg &lt;&lt; 16) | (dac &lt;&lt; 20);

		tv_dac-&gt;tv_std = radeon_atombios_get_tv_info(rdev);
	}
	return tv_dac;
}

static const char *thermal_controller_names[] = {
	"NONE",
	"LM63",
	"ADM1032",
	"ADM1030",
	"MUA6649",
	"LM64",
	"F75375",
	"ASC7512",
};

static const char *pp_lib_thermal_controller_names[] = {
	"NONE",
	"LM63",
	"ADM1032",
	"ADM1030",
	"MUA6649",
	"LM64",
	"F75375",
	"RV6xx",
	"RV770",
	"ADT7473",
};

union power_info {
	struct _ATOM_POWERPLAY_INFO info;
	struct _ATOM_POWERPLAY_INFO_V2 info_2;
	struct _ATOM_POWERPLAY_INFO_V3 info_3;
	struct _ATOM_PPLIB_POWERPLAYTABLE info_4;
};

void radeon_atombios_get_power_modes(struct radeon_device *rdev)
{
	struct radeon_mode_info *mode_info = &amp;rdev-&gt;mode_info;
	int index = GetIndexIntoMasterTable(DATA, PowerPlayInfo);
	u16 data_offset;
	u8 frev, crev;
	u32 misc, misc2 = 0, sclk, mclk;
	union power_info *power_info;
	struct _ATOM_PPLIB_NONCLOCK_INFO *non_clock_info;
	struct _ATOM_PPLIB_STATE *power_state;
	int num_modes = 0, i, j;
	int state_index = 0, mode_index = 0;
	struct radeon_i2c_bus_rec i2c_bus;

	rdev-&gt;pm.default_power_state = NULL;

	if (atom_parse_data_header(mode_info-&gt;atom_context, index, NULL,
				   &amp;frev, &amp;crev, &amp;data_offset)) {
		power_info = (union power_info *)(mode_info-&gt;atom_context-&gt;bios + data_offset);
		if (frev &lt; 4) {
			/* add the i2c bus for thermal/fan chip */
			if (power_info-&gt;info.ucOverdriveThermalController &gt; 0) {
				DRM_INFO("Possible %s thermal controller at 0x%02x\n",
					 thermal_controller_names[power_info-&gt;info.ucOverdriveThermalController],
					 power_info-&gt;info.ucOverdriveControllerAddress &gt;&gt; 1);
				i2c_bus = radeon_lookup_i2c_gpio(rdev, power_info-&gt;info.ucOverdriveI2cLine);
				rdev-&gt;pm.i2c_bus = radeon_i2c_create(rdev-&gt;ddev, &amp;i2c_bus, "Thermal");
			}
			num_modes = power_info-&gt;info.ucNumOfPowerModeEntries;
			if (num_modes &gt; ATOM_MAX_NUMBEROF_POWER_BLOCK)
				num_modes = ATOM_MAX_NUMBEROF_POWER_BLOCK;
			for (i = 0; i &lt; num_modes; i++) {
				rdev-&gt;pm.power_state[state_index].clock_info[0].voltage.type = VOLTAGE_NONE;
				switch (frev) {
				case 1:
					rdev-&gt;pm.power_state[state_index].num_clock_modes = 1;
					rdev-&gt;pm.power_state[state_index].clock_info[0].mclk =
						le16_to_cpu(power_info-&gt;info.asPowerPlayInfo[i].usMemoryClock);
					rdev-&gt;pm.power_state[state_index].clock_info[0].sclk =
						le16_to_cpu(power_info-&gt;info.asPowerPlayInfo[i].usEngineClock);
					/* skip invalid modes */
					if ((rdev-&gt;pm.power_state[state_index].clock_info[0].mclk == 0) ||
					    (rdev-&gt;pm.power_state[state_index].clock_info[0].sclk == 0))
						continue;
					/* skip overclock modes for now */
					if ((rdev-&gt;pm.power_state[state_index].clock_info[0].mclk &gt;
					     rdev-&gt;clock.default_mclk + RADEON_MODE_OVERCLOCK_MARGIN) ||
					    (rdev-&gt;pm.power_state[state_index].clock_info[0].sclk &gt;
					     rdev-&gt;clock.default_sclk + RADEON_MODE_OVERCLOCK_MARGIN))
						continue;
					rdev-&gt;pm.power_state[state_index].non_clock_info.pcie_lanes =
						power_info-&gt;info.asPowerPlayInfo[i].ucNumPciELanes;
					misc = le32_to_cpu(power_info-&gt;info.asPowerPlayInfo[i].ulMiscInfo);
					if (misc &amp; ATOM_PM_MISCINFO_VOLTAGE_DROP_SUPPORT) {
						rdev-&gt;pm.power_state[state_index].clock_info[0].voltage.type =
							VOLTAGE_GPIO;
						rdev-&gt;pm.power_state[state_index].clock_info[0].voltage.gpio =
							radeon_lookup_gpio(rdev,
							power_info-&gt;info.asPowerPlayInfo[i].ucVoltageDropIndex);
						if (misc &amp; ATOM_PM_MISCINFO_VOLTAGE_DROP_ACTIVE_HIGH)
							rdev-&gt;pm.power_state[state_index].clock_info[0].voltage.active_high =
								true;
						else
							rdev-&gt;pm.power_state[state_index].clock_info[0].voltage.active_high =
								false;
					} else if (misc &amp; ATOM_PM_MISCINFO_PROGRAM_VOLTAGE) {
						rdev-&gt;pm.power_state[state_index].clock_info[0].voltage.type =
							VOLTAGE_VDDC;
						rdev-&gt;pm.power_state[state_index].clock_info[0].voltage.vddc_id =
							power_info-&gt;info.asPowerPlayInfo[i].ucVoltageDropIndex;
					}
					/* order matters! */
					if (misc &amp; ATOM_PM_MISCINFO_POWER_SAVING_MODE)
						rdev-&gt;pm.power_state[state_index].type =
							POWER_STATE_TYPE_POWERSAVE;
					if (misc &amp; ATOM_PM_MISCINFO_DEFAULT_DC_STATE_ENTRY_TRUE)
						rdev-&gt;pm.power_state[state_index].type =
							POWER_STATE_TYPE_BATTERY;
					if (misc &amp; ATOM_PM_MISCINFO_DEFAULT_LOW_DC_STATE_ENTRY_TRUE)
						rdev-&gt;pm.power_state[state_index].type =
							POWER_STATE_TYPE_BATTERY;
					if (misc &amp; ATOM_PM_MISCINFO_LOAD_BALANCE_EN)
						rdev-&gt;pm.power_state[state_index].type =
							POWER_STATE_TYPE_BALANCED;
					if (misc &amp; ATOM_PM_MISCINFO_3D_ACCELERATION_EN)
						rdev-&gt;pm.power_state[state_index].type =
							POWER_STATE_TYPE_PERFORMANCE;
					if (misc &amp; ATOM_PM_MISCINFO_DRIVER_DEFAULT_MODE) {
						rdev-&gt;pm.power_state[state_index].type =
							POWER_STATE_TYPE_DEFAULT;
						rdev-&gt;pm.default_power_state = &amp;rdev-&gt;pm.power_state[state_index];
						rdev-&gt;pm.power_state[state_index].default_clock_mode =
							&amp;rdev-&gt;pm.power_state[state_index].clock_info[0];
					}
					state_index++;
					break;
				case 2:
					rdev-&gt;pm.power_state[state_index].num_clock_modes = 1;
					rdev-&gt;pm.power_state[state_index].clock_info[0].mclk =
						le32_to_cpu(power_info-&gt;info_2.asPowerPlayInfo[i].ulMemoryClock);
					rdev-&gt;pm.power_state[state_index].clock_info[0].sclk =
						le32_to_cpu(power_info-&gt;info_2.asPowerPlayInfo[i].ulEngineClock);
					/* skip invalid modes */
					if ((rdev-&gt;pm.power_state[state_index].clock_info[0].mclk == 0) ||
					    (rdev-&gt;pm.power_state[state_index].clock_info[0].sclk == 0))
						continue;
					/* skip overclock modes for now */
					if ((rdev-&gt;pm.power_state[state_index].clock_info[0].mclk &gt;
					     rdev-&gt;clock.default_mclk + RADEON_MODE_OVERCLOCK_MARGIN) ||
					    (rdev-&gt;pm.power_state[state_index].clock_info[0].sclk &gt;
					     rdev-&gt;clock.default_sclk + RADEON_MODE_OVERCLOCK_MARGIN))
						continue;
					rdev-&gt;pm.power_state[state_index].non_clock_info.pcie_lanes =
						power_info-&gt;info_2.asPowerPlayInfo[i].ucNumPciELanes;
					misc = le32_to_cpu(power_info-&gt;info_2.asPowerPlayInfo[i].ulMiscInfo);
					misc2 = le32_to_cpu(power_info-&gt;info_2.asPowerPlayInfo[i].ulMiscInfo2);
					if (misc &amp; ATOM_PM_MISCINFO_VOLTAGE_DROP_SUPPORT) {
						rdev-&gt;pm.power_state[state_index].clock_info[0].voltage.type =
							VOLTAGE_GPIO;
						rdev-&gt;pm.power_state[state_index].clock_info[0].voltage.gpio =
							radeon_lookup_gpio(rdev,
							power_info-&gt;info_2.asPowerPlayInfo[i].ucVoltageDropIndex);
						if (misc &amp; ATOM_PM_MISCINFO_VOLTAGE_DROP_ACTIVE_HIGH)
							rdev-&gt;pm.power_state[state_index].clock_info[0].voltage.active_high =
								true;
						else
							rdev-&gt;pm.power_state[state_index].clock_info[0].voltage.active_high =
								false;
					} else if (misc &amp; ATOM_PM_MISCINFO_PROGRAM_VOLTAGE) {
						rdev-&gt;pm.power_state[state_index].clock_info[0].voltage.type =
							VOLTAGE_VDDC;
						rdev-&gt;pm.power_state[state_index].clock_info[0].voltage.vddc_id =
							power_info-&gt;info_2.asPowerPlayInfo[i].ucVoltageDropIndex;
					}
					/* order matters! */
					if (misc &amp; ATOM_PM_MISCINFO_POWER_SAVING_MODE)
						rdev-&gt;pm.power_state[state_index].type =
							POWER_STATE_TYPE_POWERSAVE;
					if (misc &amp; ATOM_PM_MISCINFO_DEFAULT_DC_STATE_ENTRY_TRUE)
						rdev-&gt;pm.power_state[state_index].type =
							POWER_STATE_TYPE_BATTERY;
					if (misc &amp; ATOM_PM_MISCINFO_DEFAULT_LOW_DC_STATE_ENTRY_TRUE)
						rdev-&gt;pm.power_state[state_index].type =
							POWER_STATE_TYPE_BATTERY;
					if (misc &amp; ATOM_PM_MISCINFO_LOAD_BALANCE_EN)
						rdev-&gt;pm.power_state[state_index].type =
							POWER_STATE_TYPE_BALANCED;
					if (misc &amp; ATOM_PM_MISCINFO_3D_ACCELERATION_EN)
						rdev-&gt;pm.power_state[state_index].type =
							POWER_STATE_TYPE_PERFORMANCE;
					if (misc2 &amp; ATOM_PM_MISCINFO2_SYSTEM_AC_LITE_MODE)
						rdev-&gt;pm.power_state[state_index].type =
							POWER_STATE_TYPE_BALANCED;
					if (misc &amp; ATOM_PM_MISCINFO_DRIVER_DEFAULT_MODE) {
						rdev-&gt;pm.power_state[state_index].type =
							POWER_STATE_TYPE_DEFAULT;
						rdev-&gt;pm.default_power_state = &amp;rdev-&gt;pm.power_state[state_index];
						rdev-&gt;pm.power_state[state_index].default_clock_mode =
							&amp;rdev-&gt;pm.power_state[state_index].clock_info[0];
					}
					state_index++;
					break;
				case 3:
					rdev-&gt;pm.power_state[state_index].num_clock_modes = 1;
					rdev-&gt;pm.power_state[state_index].clock_info[0].mclk =
						le32_to_cpu(power_info-&gt;info_3.asPowerPlayInfo[i].ulMemoryClock);
					rdev-&gt;pm.power_state[state_index].clock_info[0].sclk =
						le32_to_cpu(power_info-&gt;info_3.asPowerPlayInfo[i].ulEngineClock);
					/* skip invalid modes */
					if ((rdev-&gt;pm.power_state[state_index].clock_info[0].mclk == 0) ||
					    (rdev-&gt;pm.power_state[state_index].clock_info[0].sclk == 0))
						continue;
					/* skip overclock modes for now */
					if ((rdev-&gt;pm.power_state[state_index].clock_info[0].mclk &gt;
					     rdev-&gt;clock.default_mclk + RADEON_MODE_OVERCLOCK_MARGIN) ||
					    (rdev-&gt;pm.power_state[state_index].clock_info[0].sclk &gt;
					     rdev-&gt;clock.default_sclk + RADEON_MODE_OVERCLOCK_MARGIN))
						continue;
					rdev-&gt;pm.power_state[state_index].non_clock_info.pcie_lanes =
						power_info-&gt;info_3.asPowerPlayInfo[i].ucNumPciELanes;
					misc = le32_to_cpu(power_info-&gt;info_3.asPowerPlayInfo[i].ulMiscInfo);
					misc2 = le32_to_cpu(power_info-&gt;info_3.asPowerPlayInfo[i].ulMiscInfo2);
					if (misc &amp; ATOM_PM_MISCINFO_VOLTAGE_DROP_SUPPORT) {
						rdev-&gt;pm.power_state[state_index].clock_info[0].voltage.type =
							VOLTAGE_GPIO;
						rdev-&gt;pm.power_state[state_index].clock_info[0].voltage.gpio =
							radeon_lookup_gpio(rdev,
							power_info-&gt;info_3.asPowerPlayInfo[i].ucVoltageDropIndex);
						if (misc &amp; ATOM_PM_MISCINFO_VOLTAGE_DROP_ACTIVE_HIGH)
							rdev-&gt;pm.power_state[state_index].clock_info[0].voltage.active_high =
								true;
						else
							rdev-&gt;pm.power_state[state_index].clock_info[0].voltage.active_high =
								false;
					} else if (misc &amp; ATOM_PM_MISCINFO_PROGRAM_VOLTAGE) {
						rdev-&gt;pm.power_state[state_index].clock_info[0].voltage.type =
							VOLTAGE_VDDC;
						rdev-&gt;pm.power_state[state_index].clock_info[0].voltage.vddc_id =
							power_info-&gt;info_3.asPowerPlayInfo[i].ucVoltageDropIndex;
						if (misc2 &amp; ATOM_PM_MISCINFO2_VDDCI_DYNAMIC_VOLTAGE_EN) {
							rdev-&gt;pm.power_state[state_index].clock_info[0].voltage.vddci_enabled =
								true;
							rdev-&gt;pm.power_state[state_index].clock_info[0].voltage.vddci_id =
							power_info-&gt;info_3.asPowerPlayInfo[i].ucVDDCI_VoltageDropIndex;
						}
					}
					/* order matters! */
					if (misc &amp; ATOM_PM_MISCINFO_POWER_SAVING_MODE)
						rdev-&gt;pm.power_state[state_index].type =
							POWER_STATE_TYPE_POWERSAVE;
					if (misc &amp; ATOM_PM_MISCINFO_DEFAULT_DC_STATE_ENTRY_TRUE)
						rdev-&gt;pm.power_state[state_index].type =
							POWER_STATE_TYPE_BATTERY;
					if (misc &amp; ATOM_PM_MISCINFO_DEFAULT_LOW_DC_STATE_ENTRY_TRUE)
						rdev-&gt;pm.power_state[state_index].type =
							POWER_STATE_TYPE_BATTERY;
					if (misc &amp; ATOM_PM_MISCINFO_LOAD_BALANCE_EN)
						rdev-&gt;pm.power_state[state_index].type =
							POWER_STATE_TYPE_BALANCED;
					if (misc &amp; ATOM_PM_MISCINFO_3D_ACCELERATION_EN)
						rdev-&gt;pm.power_state[state_index].type =
							POWER_STATE_TYPE_PERFORMANCE;
					if (misc2 &amp; ATOM_PM_MISCINFO2_SYSTEM_AC_LITE_MODE)
						rdev-&gt;pm.power_state[state_index].type =
							POWER_STATE_TYPE_BALANCED;
					if (misc &amp; ATOM_PM_MISCINFO_DRIVER_DEFAULT_MODE) {
						rdev-&gt;pm.power_state[state_index].type =
							POWER_STATE_TYPE_DEFAULT;
						rdev-&gt;pm.default_power_state = &amp;rdev-&gt;pm.power_state[state_index];
						rdev-&gt;pm.power_state[state_index].default_clock_mode =
							&amp;rdev-&gt;pm.power_state[state_index].clock_info[0];
					}
					state_index++;
					break;
				}
			}
		} else if (frev == 4) {
			/* add the i2c bus for thermal/fan chip */
			/* no support for internal controller yet */
			if (power_info-&gt;info_4.sThermalController.ucType &gt; 0) {
				if ((power_info-&gt;info_4.sThermalController.ucType == ATOM_PP_THERMALCONTROLLER_RV6xx) ||
				    (power_info-&gt;info_4.sThermalController.ucType == ATOM_PP_THERMALCONTROLLER_RV770)) {
					DRM_INFO("Internal thermal controller %s fan control\n",
						 (power_info-&gt;info_4.sThermalController.ucFanParameters &amp;
						  ATOM_PP_FANPARAMETERS_NOFAN) ? "without" : "with");
				} else {
					DRM_INFO("Possible %s thermal controller at 0x%02x %s fan control\n",
						 pp_lib_thermal_controller_names[power_info-&gt;info_4.sThermalController.ucType],
						 power_info-&gt;info_4.sThermalController.ucI2cAddress &gt;&gt; 1,
						 (power_info-&gt;info_4.sThermalController.ucFanParameters &amp;
						  ATOM_PP_FANPARAMETERS_NOFAN) ? "without" : "with");
					i2c_bus = radeon_lookup_i2c_gpio(rdev, power_info-&gt;info_4.sThermalController.ucI2cLine);
					rdev-&gt;pm.i2c_bus = radeon_i2c_create(rdev-&gt;ddev, &amp;i2c_bus, "Thermal");
				}
			}
			for (i = 0; i &lt; power_info-&gt;info_4.ucNumStates; i++) {
				mode_index = 0;
				power_state = (struct _ATOM_PPLIB_STATE *)
					(mode_info-&gt;atom_context-&gt;bios +
					 data_offset +
					 le16_to_cpu(power_info-&gt;info_4.usStateArrayOffset) +
					 i * power_info-&gt;info_4.ucStateEntrySize);
				non_clock_info = (struct _ATOM_PPLIB_NONCLOCK_INFO *)
					(mode_info-&gt;atom_context-&gt;bios +
					 data_offset +
					 le16_to_cpu(power_info-&gt;info_4.usNonClockInfoArrayOffset) +
					 (power_state-&gt;ucNonClockStateIndex *
					  power_info-&gt;info_4.ucNonClockSize));
				for (j = 0; j &lt; (power_info-&gt;info_4.ucStateEntrySize - 1); j++) {
					if (rdev-&gt;flags &amp; RADEON_IS_IGP) {
						struct _ATOM_PPLIB_RS780_CLOCK_INFO *clock_info =
							(struct _ATOM_PPLIB_RS780_CLOCK_INFO *)
							(mode_info-&gt;atom_context-&gt;bios +
							 data_offset +
							 le16_to_cpu(power_info-&gt;info_4.usClockInfoArrayOffset) +
							 (power_state-&gt;ucClockStateIndices[j] *
							  power_info-&gt;info_4.ucClockInfoSize));
						sclk = le16_to_cpu(clock_info-&gt;usLowEngineClockLow);
						sclk |= clock_info-&gt;ucLowEngineClockHigh &lt;&lt; 16;
						rdev-&gt;pm.power_state[state_index].clock_info[mode_index].sclk = sclk;
						/* skip invalid modes */
						if (rdev-&gt;pm.power_state[state_index].clock_info[mode_index].sclk == 0)
							continue;
						/* skip overclock modes for now */
						if (rdev-&gt;pm.power_state[state_index].clock_info[mode_index].sclk &gt;
						    rdev-&gt;clock.default_sclk + RADEON_MODE_OVERCLOCK_MARGIN)
							continue;
						rdev-&gt;pm.power_state[state_index].clock_info[mode_index].voltage.type =
							VOLTAGE_SW;
						rdev-&gt;pm.power_state[state_index].clock_info[mode_index].voltage.voltage =
							clock_info-&gt;usVDDC;
						mode_index++;
					} else {
						struct _ATOM_PPLIB_R600_CLOCK_INFO *clock_info =
							(struct _ATOM_PPLIB_R600_CLOCK_INFO *)
							(mode_info-&gt;atom_context-&gt;bios +
							 data_offset +
							 le16_to_cpu(power_info-&gt;info_4.usClockInfoArrayOffset) +
							 (power_state-&gt;ucClockStateIndices[j] *
							  power_info-&gt;info_4.ucClockInfoSize));
						sclk = le16_to_cpu(clock_info-&gt;usEngineClockLow);
						sclk |= clock_info-&gt;ucEngineClockHigh &lt;&lt; 16;
						mclk = le16_to_cpu(clock_info-&gt;usMemoryClockLow);
						mclk |= clock_info-&gt;ucMemoryClockHigh &lt;&lt; 16;
						rdev-&gt;pm.power_state[state_index].clock_info[mode_index].mclk = mclk;
						rdev-&gt;pm.power_state[state_index].clock_info[mode_index].sclk = sclk;
						/* skip invalid modes */
						if ((rdev-&gt;pm.power_state[state_index].clock_info[mode_index].mclk == 0) ||
						    (rdev-&gt;pm.power_state[state_index].clock_info[mode_index].sclk == 0))
							continue;
						/* skip overclock modes for now */
						if ((rdev-&gt;pm.power_state[state_index].clock_info[mode_index].mclk &gt;
						     rdev-&gt;clock.default_mclk + RADEON_MODE_OVERCLOCK_MARGIN) ||
						    (rdev-&gt;pm.power_state[state_index].clock_info[mode_index].sclk &gt;
						     rdev-&gt;clock.default_sclk + RADEON_MODE_OVERCLOCK_MARGIN))
							continue;
						rdev-&gt;pm.power_state[state_index].clock_info[mode_index].voltage.type =
							VOLTAGE_SW;
						rdev-&gt;pm.power_state[state_index].clock_info[mode_index].voltage.voltage =
							clock_info-&gt;usVDDC;
						mode_index++;
					}
				}
				rdev-&gt;pm.power_state[state_index].num_clock_modes = mode_index;
				if (mode_index) {
					misc = le32_to_cpu(non_clock_info-&gt;ulCapsAndSettings);
					misc2 = le16_to_cpu(non_clock_info-&gt;usClassification);
					rdev-&gt;pm.power_state[state_index].non_clock_info.pcie_lanes =
						((misc &amp; ATOM_PPLIB_PCIE_LINK_WIDTH_MASK) &gt;&gt;
						ATOM_PPLIB_PCIE_LINK_WIDTH_SHIFT) + 1;
					switch (misc2 &amp; ATOM_PPLIB_CLASSIFICATION_UI_MASK) {
					case ATOM_PPLIB_CLASSIFICATION_UI_BATTERY:
						rdev-&gt;pm.power_state[state_index].type =
							POWER_STATE_TYPE_BATTERY;
						break;
					case ATOM_PPLIB_CLASSIFICATION_UI_BALANCED:
						rdev-&gt;pm.power_state[state_index].type =
							POWER_STATE_TYPE_BALANCED;
						break;
					case ATOM_PPLIB_CLASSIFICATION_UI_PERFORMANCE:
						rdev-&gt;pm.power_state[state_index].type =
							POWER_STATE_TYPE_PERFORMANCE;
						break;
					}
					if (misc2 &amp; ATOM_PPLIB_CLASSIFICATION_BOOT) {
						rdev-&gt;pm.power_state[state_index].type =
							POWER_STATE_TYPE_DEFAULT;
						rdev-&gt;pm.default_power_state = &amp;rdev-&gt;pm.power_state[state_index];
						rdev-&gt;pm.power_state[state_index].default_clock_mode =
							&amp;rdev-&gt;pm.power_state[state_index].clock_info[mode_index - 1];
					}
					state_index++;
				}
			}
		}
	} else {
		/* XXX figure out some good default low power mode for cards w/out power tables */
	}

	if (rdev-&gt;pm.default_power_state == NULL) {
		/* add the default mode */
		rdev-&gt;pm.power_state[state_index].type =
			POWER_STATE_TYPE_DEFAULT;
		rdev-&gt;pm.power_state[state_index].num_clock_modes = 1;
		rdev-&gt;pm.power_state[state_index].clock_info[0].mclk = rdev-&gt;clock.default_mclk;
		rdev-&gt;pm.power_state[state_index].clock_info[0].sclk = rdev-&gt;clock.default_sclk;
		rdev-&gt;pm.power_state[state_index].default_clock_mode =
			&amp;rdev-&gt;pm.power_state[state_index].clock_info[0];
		rdev-&gt;pm.power_state[state_index].clock_info[0].voltage.type = VOLTAGE_NONE;
		if (rdev-&gt;asic-&gt;get_pcie_lanes)
			rdev-&gt;pm.power_state[state_index].non_clock_info.pcie_lanes = radeon_get_pcie_lanes(rdev);
		else
			rdev-&gt;pm.power_state[state_index].non_clock_info.pcie_lanes = 16;
		rdev-&gt;pm.default_power_state = &amp;rdev-&gt;pm.power_state[state_index];
		state_index++;
	}
	rdev-&gt;pm.num_power_states = state_index;

	rdev-&gt;pm.current_power_state = rdev-&gt;pm.default_power_state;
	rdev-&gt;pm.current_clock_mode =
		rdev-&gt;pm.default_power_state-&gt;default_clock_mode;
}

void radeon_atom_set_clock_gating(struct radeon_device *rdev, int enable)
{
	DYNAMIC_CLOCK_GATING_PS_ALLOCATION args;
	int index = GetIndexIntoMasterTable(COMMAND, DynamicClockGating);

	args.ucEnable = enable;

	atom_execute_table(rdev-&gt;mode_info.atom_context, index, (uint32_t *)&amp;args);
}

uint32_t radeon_atom_get_engine_clock(struct radeon_device *rdev)
{
	GET_ENGINE_CLOCK_PS_ALLOCATION args;
	int index = GetIndexIntoMasterTable(COMMAND, GetEngineClock);

	atom_execute_table(rdev-&gt;mode_info.atom_context, index, (uint32_t *)&amp;args);
	return args.ulReturnEngineClock;
}

uint32_t radeon_atom_get_memory_clock(struct radeon_device *rdev)
{
	GET_MEMORY_CLOCK_PS_ALLOCATION args;
	int index = GetIndexIntoMasterTable(COMMAND, GetMemoryClock);

	atom_execute_table(rdev-&gt;mode_info.atom_context, index, (uint32_t *)&amp;args);
	return args.ulReturnMemoryClock;
}

void radeon_atom_set_engine_clock(struct radeon_device *rdev,
				  uint32_t eng_clock)
{
	SET_ENGINE_CLOCK_PS_ALLOCATION args;
	int index = GetIndexIntoMasterTable(COMMAND, SetEngineClock);

	args.ulTargetEngineClock = eng_clock;	/* 10 khz */

	atom_execute_table(rdev-&gt;mode_info.atom_context, index, (uint32_t *)&amp;args);
}

void radeon_atom_set_memory_clock(struct radeon_device *rdev,
				  uint32_t mem_clock)
{
	SET_MEMORY_CLOCK_PS_ALLOCATION args;
	int index = GetIndexIntoMasterTable(COMMAND, SetMemoryClock);

	if (rdev-&gt;flags &amp; RADEON_IS_IGP)
		return;

	args.ulTargetMemoryClock = mem_clock;	/* 10 khz */

	atom_execute_table(rdev-&gt;mode_info.atom_context, index, (uint32_t *)&amp;args);
}

void radeon_atom_initialize_bios_scratch_regs(struct drm_device *dev)
{
	struct radeon_device *rdev = dev-&gt;dev_private;
	uint32_t bios_2_scratch, bios_6_scratch;

	if (rdev-&gt;family &gt;= CHIP_R600) {
		bios_2_scratch = RREG32(R600_BIOS_2_SCRATCH);
		bios_6_scratch = RREG32(R600_BIOS_6_SCRATCH);
	} else {
		bios_2_scratch = RREG32(RADEON_BIOS_2_SCRATCH);
		bios_6_scratch = RREG32(RADEON_BIOS_6_SCRATCH);
	}

	/* let the bios control the backlight */
	bios_2_scratch &amp;= ~ATOM_S2_VRI_BRIGHT_ENABLE;

	/* tell the bios not to handle mode switching */
	bios_6_scratch |= (ATOM_S6_ACC_BLOCK_DISPLAY_SWITCH | ATOM_S6_ACC_MODE);

	if (rdev-&gt;family &gt;= CHIP_R600) {
		WREG32(R600_BIOS_2_SCRATCH, bios_2_scratch);
		WREG32(R600_BIOS_6_SCRATCH, bios_6_scratch);
	} else {
		WREG32(RADEON_BIOS_2_SCRATCH, bios_2_scratch);
		WREG32(RADEON_BIOS_6_SCRATCH, bios_6_scratch);
	}

}

void radeon_save_bios_scratch_regs(struct radeon_device *rdev)
{
	uint32_t scratch_reg;
	int i;

	if (rdev-&gt;family &gt;= CHIP_R600)
		scratch_reg = R600_BIOS_0_SCRATCH;
	else
		scratch_reg = RADEON_BIOS_0_SCRATCH;

	for (i = 0; i &lt; RADEON_BIOS_NUM_SCRATCH; i++)
		rdev-&gt;bios_scratch[i] = RREG32(scratch_reg + (i * 4));
}

void radeon_restore_bios_scratch_regs(struct radeon_device *rdev)
{
	uint32_t scratch_reg;
	int i;

	if (rdev-&gt;family &gt;= CHIP_R600)
		scratch_reg = R600_BIOS_0_SCRATCH;
	else
		scratch_reg = RADEON_BIOS_0_SCRATCH;

	for (i = 0; i &lt; RADEON_BIOS_NUM_SCRATCH; i++)
		WREG32(scratch_reg + (i * 4), rdev-&gt;bios_scratch[i]);
}

void radeon_atom_output_lock(struct drm_encoder *encoder, bool lock)
{
	struct drm_device *dev = encoder-&gt;dev;
	struct radeon_device *rdev = dev-&gt;dev_private;
	uint32_t bios_6_scratch;

	if (rdev-&gt;family &gt;= CHIP_R600)
		bios_6_scratch = RREG32(R600_BIOS_6_SCRATCH);
	else
		bios_6_scratch = RREG32(RADEON_BIOS_6_SCRATCH);

	if (lock)
		bios_6_scratch |= ATOM_S6_CRITICAL_STATE;
	else
		bios_6_scratch &amp;= ~ATOM_S6_CRITICAL_STATE;

	if (rdev-&gt;family &gt;= CHIP_R600)
		WREG32(R600_BIOS_6_SCRATCH, bios_6_scratch);
	else
		WREG32(RADEON_BIOS_6_SCRATCH, bios_6_scratch);
}

/* at some point we may want to break this out into individual functions */
void
radeon_atombios_connected_scratch_regs(struct drm_connector *connector,
				       struct drm_encoder *encoder,
				       bool connected)
{
	struct drm_device *dev = connector-&gt;dev;
	struct radeon_device *rdev = dev-&gt;dev_private;
	struct radeon_connector *radeon_connector =
	    to_radeon_connector(connector);
	struct radeon_encoder *radeon_encoder = to_radeon_encoder(encoder);
	uint32_t bios_0_scratch, bios_3_scratch, bios_6_scratch;

	if (rdev-&gt;family &gt;= CHIP_R600) {
		bios_0_scratch = RREG32(R600_BIOS_0_SCRATCH);
		bios_3_scratch = RREG32(R600_BIOS_3_SCRATCH);
		bios_6_scratch = RREG32(R600_BIOS_6_SCRATCH);
	} else {
		bios_0_scratch = RREG32(RADEON_BIOS_0_SCRATCH);
		bios_3_scratch = RREG32(RADEON_BIOS_3_SCRATCH);
		bios_6_scratch = RREG32(RADEON_BIOS_6_SCRATCH);
	}

	if ((radeon_encoder-&gt;devices &amp; ATOM_DEVICE_TV1_SUPPORT) &amp;&amp;
	    (radeon_connector-&gt;devices &amp; ATOM_DEVICE_TV1_SUPPORT)) {
		if (connected) {
			DRM_DEBUG("TV1 connected\n");
			bios_3_scratch |= ATOM_S3_TV1_ACTIVE;
			bios_6_scratch |= ATOM_S6_ACC_REQ_TV1;
		} else {
			DRM_DEBUG("TV1 disconnected\n");
			bios_0_scratch &amp;= ~ATOM_S0_TV1_MASK;
			bios_3_scratch &amp;= ~ATOM_S3_TV1_ACTIVE;
			bios_6_scratch &amp;= ~ATOM_S6_ACC_REQ_TV1;
		}
	}
	if ((radeon_encoder-&gt;devices &amp; ATOM_DEVICE_CV_SUPPORT) &amp;&amp;
	    (radeon_connector-&gt;devices &amp; ATOM_DEVICE_CV_SUPPORT)) {
		if (connected) {
			DRM_DEBUG("CV connected\n");
			bios_3_scratch |= ATOM_S3_CV_ACTIVE;
			bios_6_scratch |= ATOM_S6_ACC_REQ_CV;
		} else {
			DRM_DEBUG("CV disconnected\n");
			bios_0_scratch &amp;= ~ATOM_S0_CV_MASK;
			bios_3_scratch &amp;= ~ATOM_S3_CV_ACTIVE;
			bios_6_scratch &amp;= ~ATOM_S6_ACC_REQ_CV;
		}
	}
	if ((radeon_encoder-&gt;devices &amp; ATOM_DEVICE_LCD1_SUPPORT) &amp;&amp;
	    (radeon_connector-&gt;devices &amp; ATOM_DEVICE_LCD1_SUPPORT)) {
		if (connected) {
			DRM_DEBUG("LCD1 connected\n");
			bios_0_scratch |= ATOM_S0_LCD1;
			bios_3_scratch |= ATOM_S3_LCD1_ACTIVE;
			bios_6_scratch |= ATOM_S6_ACC_REQ_LCD1;
		} else {
			DRM_DEBUG("LCD1 disconnected\n");
			bios_0_scratch &amp;= ~ATOM_S0_LCD1;
			bios_3_scratch &amp;= ~ATOM_S3_LCD1_ACTIVE;
			bios_6_scratch &amp;= ~ATOM_S6_ACC_REQ_LCD1;
		}
	}
	if ((radeon_encoder-&gt;devices &amp; ATOM_DEVICE_CRT1_SUPPORT) &amp;&amp;
	    (radeon_connector-&gt;devices &amp; ATOM_DEVICE_CRT1_SUPPORT)) {
		if (connected) {
			DRM_DEBUG("CRT1 connected\n");
			bios_0_scratch |= ATOM_S0_CRT1_COLOR;
			bios_3_scratch |= ATOM_S3_CRT1_ACTIVE;
			bios_6_scratch |= ATOM_S6_ACC_REQ_CRT1;
		} else {
			DRM_DEBUG("CRT1 disconnected\n");
			bios_0_scratch &amp;= ~ATOM_S0_CRT1_MASK;
			bios_3_scratch &amp;= ~ATOM_S3_CRT1_ACTIVE;
			bios_6_scratch &amp;= ~ATOM_S6_ACC_REQ_CRT1;
		}
	}
	if ((radeon_encoder-&gt;devices &amp; ATOM_DEVICE_CRT2_SUPPORT) &amp;&amp;
	    (radeon_connector-&gt;devices &amp; ATOM_DEVICE_CRT2_SUPPORT)) {
		if (connected) {
			DRM_DEBUG("CRT2 connected\n");
			bios_0_scratch |= ATOM_S0_CRT2_COLOR;
			bios_3_scratch |= ATOM_S3_CRT2_ACTIVE;
			bios_6_scratch |= ATOM_S6_ACC_REQ_CRT2;
		} else {
			DRM_DEBUG("CRT2 disconnected\n");
			bios_0_scratch &amp;= ~ATOM_S0_CRT2_MASK;
			bios_3_scratch &amp;= ~ATOM_S3_CRT2_ACTIVE;
			bios_6_scratch &amp;= ~ATOM_S6_ACC_REQ_CRT2;
		}
	}
	if ((radeon_encoder-&gt;devices &amp; ATOM_DEVICE_DFP1_SUPPORT) &amp;&amp;
	    (radeon_connector-&gt;devices &amp; ATOM_DEVICE_DFP1_SUPPORT)) {
		if (connected) {
			DRM_DEBUG("DFP1 connected\n");
			bios_0_scratch |= ATOM_S0_DFP1;
			bios_3_scratch |= ATOM_S3_DFP1_ACTIVE;
			bios_6_scratch |= ATOM_S6_ACC_REQ_DFP1;
		} else {
			DRM_DEBUG("DFP1 disconnected\n");
			bios_0_scratch &amp;= ~ATOM_S0_DFP1;
			bios_3_scratch &amp;= ~ATOM_S3_DFP1_ACTIVE;
			bios_6_scratch &amp;= ~ATOM_S6_ACC_REQ_DFP1;
		}
	}
	if ((radeon_encoder-&gt;devices &amp; ATOM_DEVICE_DFP2_SUPPORT) &amp;&amp;
	    (radeon_connector-&gt;devices &amp; ATOM_DEVICE_DFP2_SUPPORT)) {
		if (connected) {
			DRM_DEBUG("DFP2 connected\n");
			bios_0_scratch |= ATOM_S0_DFP2;
			bios_3_scratch |= ATOM_S3_DFP2_ACTIVE;
			bios_6_scratch |= ATOM_S6_ACC_REQ_DFP2;
		} else {
			DRM_DEBUG("DFP2 disconnected\n");
			bios_0_scratch &amp;= ~ATOM_S0_DFP2;
			bios_3_scratch &amp;= ~ATOM_S3_DFP2_ACTIVE;
			bios_6_scratch &amp;= ~ATOM_S6_ACC_REQ_DFP2;
		}
	}
	if ((radeon_encoder-&gt;devices &amp; ATOM_DEVICE_DFP3_SUPPORT) &amp;&amp;
	    (radeon_connector-&gt;devices &amp; ATOM_DEVICE_DFP3_SUPPORT)) {
		if (connected) {
			DRM_DEBUG("DFP3 connected\n");
			bios_0_scratch |= ATOM_S0_DFP3;
			bios_3_scratch |= ATOM_S3_DFP3_ACTIVE;
			bios_6_scratch |= ATOM_S6_ACC_REQ_DFP3;
		} else {
			DRM_DEBUG("DFP3 disconnected\n");
			bios_0_scratch &amp;= ~ATOM_S0_DFP3;
			bios_3_scratch &amp;= ~ATOM_S3_DFP3_ACTIVE;
			bios_6_scratch &amp;= ~ATOM_S6_ACC_REQ_DFP3;
		}
	}
	if ((radeon_encoder-&gt;devices &amp; ATOM_DEVICE_DFP4_SUPPORT) &amp;&amp;
	    (radeon_connector-&gt;devices &amp; ATOM_DEVICE_DFP4_SUPPORT)) {
		if (connected) {
			DRM_DEBUG("DFP4 connected\n");
			bios_0_scratch |= ATOM_S0_DFP4;
			bios_3_scratch |= ATOM_S3_DFP4_ACTIVE;
			bios_6_scratch |= ATOM_S6_ACC_REQ_DFP4;
		} else {
			DRM_DEBUG("DFP4 disconnected\n");
			bios_0_scratch &amp;= ~ATOM_S0_DFP4;
			bios_3_scratch &amp;= ~ATOM_S3_DFP4_ACTIVE;
			bios_6_scratch &amp;= ~ATOM_S6_ACC_REQ_DFP4;
		}
	}
	if ((radeon_encoder-&gt;devices &amp; ATOM_DEVICE_DFP5_SUPPORT) &amp;&amp;
	    (radeon_connector-&gt;devices &amp; ATOM_DEVICE_DFP5_SUPPORT)) {
		if (connected) {
			DRM_DEBUG("DFP5 connected\n");
			bios_0_scratch |= ATOM_S0_DFP5;
			bios_3_scratch |= ATOM_S3_DFP5_ACTIVE;
			bios_6_scratch |= ATOM_S6_ACC_REQ_DFP5;
		} else {
			DRM_DEBUG("DFP5 disconnected\n");
			bios_0_scratch &amp;= ~ATOM_S0_DFP5;
			bios_3_scratch &amp;= ~ATOM_S3_DFP5_ACTIVE;
			bios_6_scratch &amp;= ~ATOM_S6_ACC_REQ_DFP5;
		}
	}

	if (rdev-&gt;family &gt;= CHIP_R600) {
		WREG32(R600_BIOS_0_SCRATCH, bios_0_scratch);
		WREG32(R600_BIOS_3_SCRATCH, bios_3_scratch);
		WREG32(R600_BIOS_6_SCRATCH, bios_6_scratch);
	} else {
		WREG32(RADEON_BIOS_0_SCRATCH, bios_0_scratch);
		WREG32(RADEON_BIOS_3_SCRATCH, bios_3_scratch);
		WREG32(RADEON_BIOS_6_SCRATCH, bios_6_scratch);
	}
}

void
radeon_atombios_encoder_crtc_scratch_regs(struct drm_encoder *encoder, int crtc)
{
	struct drm_device *dev = encoder-&gt;dev;
	struct radeon_device *rdev = dev-&gt;dev_private;
	struct radeon_encoder *radeon_encoder = to_radeon_encoder(encoder);
	uint32_t bios_3_scratch;

	if (rdev-&gt;family &gt;= CHIP_R600)
		bios_3_scratch = RREG32(R600_BIOS_3_SCRATCH);
	else
		bios_3_scratch = RREG32(RADEON_BIOS_3_SCRATCH);

	if (radeon_encoder-&gt;devices &amp; ATOM_DEVICE_TV1_SUPPORT) {
		bios_3_scratch &amp;= ~ATOM_S3_TV1_CRTC_ACTIVE;
		bios_3_scratch |= (crtc &lt;&lt; 18);
	}
	if (radeon_encoder-&gt;devices &amp; ATOM_DEVICE_CV_SUPPORT) {
		bios_3_scratch &amp;= ~ATOM_S3_CV_CRTC_ACTIVE;
		bios_3_scratch |= (crtc &lt;&lt; 24);
	}
	if (radeon_encoder-&gt;devices &amp; ATOM_DEVICE_CRT1_SUPPORT) {
		bios_3_scratch &amp;= ~ATOM_S3_CRT1_CRTC_ACTIVE;
		bios_3_scratch |= (crtc &lt;&lt; 16);
	}
	if (radeon_encoder-&gt;devices &amp; ATOM_DEVICE_CRT2_SUPPORT) {
		bios_3_scratch &amp;= ~ATOM_S3_CRT2_CRTC_ACTIVE;
		bios_3_scratch |= (crtc &lt;&lt; 20);
	}
	if (radeon_encoder-&gt;devices &amp; ATOM_DEVICE_LCD1_SUPPORT) {
		bios_3_scratch &amp;= ~ATOM_S3_LCD1_CRTC_ACTIVE;
		bios_3_scratch |= (crtc &lt;&lt; 17);
	}
	if (radeon_encoder-&gt;devices &amp; ATOM_DEVICE_DFP1_SUPPORT) {
		bios_3_scratch &amp;= ~ATOM_S3_DFP1_CRTC_ACTIVE;
		bios_3_scratch |= (crtc &lt;&lt; 19);
	}
	if (radeon_encoder-&gt;devices &amp; ATOM_DEVICE_DFP2_SUPPORT) {
		bios_3_scratch &amp;= ~ATOM_S3_DFP2_CRTC_ACTIVE;
		bios_3_scratch |= (crtc &lt;&lt; 23);
	}
	if (radeon_encoder-&gt;devices &amp; ATOM_DEVICE_DFP3_SUPPORT) {
		bios_3_scratch &amp;= ~ATOM_S3_DFP3_CRTC_ACTIVE;
		bios_3_scratch |= (crtc &lt;&lt; 25);
	}

	if (rdev-&gt;family &gt;= CHIP_R600)
		WREG32(R600_BIOS_3_SCRATCH, bios_3_scratch);
	else
		WREG32(RADEON_BIOS_3_SCRATCH, bios_3_scratch);
}

void
radeon_atombios_encoder_dpms_scratch_regs(struct drm_encoder *encoder, bool on)
{
	struct drm_device *dev = encoder-&gt;dev;
	struct radeon_device *rdev = dev-&gt;dev_private;
	struct radeon_encoder *radeon_encoder = to_radeon_encoder(encoder);
	uint32_t bios_2_scratch;

	if (rdev-&gt;family &gt;= CHIP_R600)
		bios_2_scratch = RREG32(R600_BIOS_2_SCRATCH);
	else
		bios_2_scratch = RREG32(RADEON_BIOS_2_SCRATCH);

	if (radeon_encoder-&gt;devices &amp; ATOM_DEVICE_TV1_SUPPORT) {
		if (on)
			bios_2_scratch &amp;= ~ATOM_S2_TV1_DPMS_STATE;
		else
			bios_2_scratch |= ATOM_S2_TV1_DPMS_STATE;
	}
	if (radeon_encoder-&gt;devices &amp; ATOM_DEVICE_CV_SUPPORT) {
		if (on)
			bios_2_scratch &amp;= ~ATOM_S2_CV_DPMS_STATE;
		else
			bios_2_scratch |= ATOM_S2_CV_DPMS_STATE;
	}
	if (radeon_encoder-&gt;devices &amp; ATOM_DEVICE_CRT1_SUPPORT) {
		if (on)
			bios_2_scratch &amp;= ~ATOM_S2_CRT1_DPMS_STATE;
		else
			bios_2_scratch |= ATOM_S2_CRT1_DPMS_STATE;
	}
	if (radeon_encoder-&gt;devices &amp; ATOM_DEVICE_CRT2_SUPPORT) {
		if (on)
			bios_2_scratch &amp;= ~ATOM_S2_CRT2_DPMS_STATE;
		else
			bios_2_scratch |= ATOM_S2_CRT2_DPMS_STATE;
	}
	if (radeon_encoder-&gt;devices &amp; ATOM_DEVICE_LCD1_SUPPORT) {
		if (on)
			bios_2_scratch &amp;= ~ATOM_S2_LCD1_DPMS_STATE;
		else
			bios_2_scratch |= ATOM_S2_LCD1_DPMS_STATE;
	}
	if (radeon_encoder-&gt;devices &amp; ATOM_DEVICE_DFP1_SUPPORT) {
		if (on)
			bios_2_scratch &amp;= ~ATOM_S2_DFP1_DPMS_STATE;
		else
			bios_2_scratch |= ATOM_S2_DFP1_DPMS_STATE;
	}
	if (radeon_encoder-&gt;devices &amp; ATOM_DEVICE_DFP2_SUPPORT) {
		if (on)
			bios_2_scratch &amp;= ~ATOM_S2_DFP2_DPMS_STATE;
		else
			bios_2_scratch |= ATOM_S2_DFP2_DPMS_STATE;
	}
	if (radeon_encoder-&gt;devices &amp; ATOM_DEVICE_DFP3_SUPPORT) {
		if (on)
			bios_2_scratch &amp;= ~ATOM_S2_DFP3_DPMS_STATE;
		else
			bios_2_scratch |= ATOM_S2_DFP3_DPMS_STATE;
	}
	if (radeon_encoder-&gt;devices &amp; ATOM_DEVICE_DFP4_SUPPORT) {
		if (on)
			bios_2_scratch &amp;= ~ATOM_S2_DFP4_DPMS_STATE;
		else
			bios_2_scratch |= ATOM_S2_DFP4_DPMS_STATE;
	}
	if (radeon_encoder-&gt;devices &amp; ATOM_DEVICE_DFP5_SUPPORT) {
		if (on)
			bios_2_scratch &amp;= ~ATOM_S2_DFP5_DPMS_STATE;
		else
			bios_2_scratch |= ATOM_S2_DFP5_DPMS_STATE;
	}

	if (rdev-&gt;family &gt;= CHIP_R600)
		WREG32(R600_BIOS_2_SCRATCH, bios_2_scratch);
	else
		WREG32(RADEON_BIOS_2_SCRATCH, bios_2_scratch);
}
